from typing import Union

import torch

from esp_ppq.api.setting import QuantizationSetting
from esp_ppq.core import (
    PASSIVE_OPERATIONS,
    OperationQuantizationConfig,
    QuantizationPolicy,
    QuantizationProperty,
    QuantizationStates,
    QuantizationVisibility,
    RoundingPolicy,
    TargetPlatform,
)
from esp_ppq.IR import BaseGraph, Operation
from esp_ppq.quantization.optim import PPLDSPTIReCalibrationPass, QuantizationOptimizationPipeline

from .base import BaseQuantizer


class PPL_DSP_Quantizer(BaseQuantizer):
    def __init__(self, graph: BaseGraph) -> Union[torch.Tensor, list, dict]:
        super().__init__(graph=graph)
        self._num_of_bits = 8
        self._quant_min = 0
        self._quant_max = int(pow(2, self._num_of_bits) - 1)

    def init_quantize_config(self, operation: Operation) -> OperationQuantizationConfig:
        base_quant_config = self.create_default_quant_config(
            op=operation,
            num_of_bits=self._num_of_bits,
            exponent_bits=0,
            quant_max=self._quant_max,
            quant_min=self._quant_min,
            observer_algorithm='percentile',
            policy=self.quantize_policy,
            rounding=self.rounding_policy,
        )

        if operation.type in {'Conv', 'ConvTranspose', 'Gemm'}:
            # if operation has bias
            if operation.num_of_input == 3:
                bias_config = base_quant_config.input_quantization_config[-1]
                # bias should be quantized with 32 bits
                # in python3, int indicates long long in C++
                # so that it has enough precision to represent a number like 2^32
                # however, it may cause a scale underflow
                # here we give bias a 30 bits precision, which is pettery enough in all cases
                bias_config.num_of_bits = 30
                bias_config.quant_max = int(pow(2, 30 - 1) - 1)
                bias_config.quant_min = -int(pow(2, 30 - 1))
                bias_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_TENSOR
                )
                bias_config.state = QuantizationStates.PASSIVE_INIT
                bias_config.visibility = QuantizationVisibility.INTERNAL

            for config in base_quant_config.input_quantization_config[1:]:
                config.observer_algorithm = 'minmax'

        if operation.type in {'Clip', 'Pad'}:
            # Clip min, max must be PASSIVE INIT state
            # Padding value must be PASSIVE INIT state
            # They will be processed with PassiveParameterQuantizationPass
            if operation.type == 'Clip':
                if operation.num_of_input > 1:
                    min_cfg = base_quant_config.input_quantization_config[1]
                    max_cfg = base_quant_config.input_quantization_config[2]

                    min_cfg.state = QuantizationStates.PASSIVE_INIT
                    max_cfg.state = QuantizationStates.PASSIVE_INIT

                    min_cfg.visibility = QuantizationVisibility.INTERNAL
                    max_cfg.visibility = QuantizationVisibility.INTERNAL

            if operation.type == 'Pad':
                if operation.num_of_input > 2:
                    constant_cfg = base_quant_config.input_quantization_config[-1]
                    constant_cfg = QuantizationStates.PASSIVE_INIT
                    constant_cfg.visiblity = QuantizationVisibility.INTERNAL

        if operation.type in PASSIVE_OPERATIONS:
            # Those op are not active op.
            base_quant_config.is_active_quant_op = False
        return base_quant_config

    @property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.PPL_DSP_INT8

    @property
    def default_platform(self) -> TargetPlatform:
        return TargetPlatform.FP32

    @property
    def quant_operation_types(self) -> set:
        return {
            'Conv',
            'ConvTranspose',
            'Gemm',
            'Relu',
            'PRelu',
            'Clip',
            'Pad',
            'Resize',
            'MaxPool',
            'AveragePool',
            'GlobalMaxPool',
            'GlobalAveragePool',
            'Softmax',
            'Mul',
            'Add',
            'Max',
            'Sub',
            'Div',
            'Reshape',
            'LeakyRelu',
            'Concat',
            'Sigmoid',
            'Slice',
            'Interp',
            'ReduceMean',
            'Flatten',
            'Scale',
        }

    @property
    def quantize_policy(self) -> QuantizationPolicy:
        return QuantizationPolicy(
            QuantizationProperty.ASYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_TENSOR
        )

    @property
    def rounding_policy(self) -> RoundingPolicy:
        return RoundingPolicy.ROUND_HALF_EVEN

    @property
    def activation_fusion_types(self) -> set:
        return {'Relu', 'Clip'}


class PPL_DSP_TI_Quantizer(PPL_DSP_Quantizer):
    def __init__(self, graph: BaseGraph) -> Union[torch.Tensor, list, dict]:
        super().__init__(graph)
        self._num_of_bits = 8
        self._quant_min = -int(pow(2, self._num_of_bits - 1))
        self._quant_max = int(pow(2, self._num_of_bits - 1)) - 1

    def build_quant_pipeline(self, setting: QuantizationSetting) -> QuantizationOptimizationPipeline:
        pipeline = super().build_quant_pipeline(setting)
        pipeline.append_optimization_to_pipeline(PPLDSPTIReCalibrationPass())
        return pipeline

    def init_quantize_config(self, operation: Operation) -> OperationQuantizationConfig:
        base_quant_config = self.create_default_quant_config(
            policy=self.quantize_policy,
            rounding=self.rounding_policy,
            op=operation,
            num_of_bits=self._num_of_bits,
            exponent_bits=0,
            quant_max=self._quant_max,
            quant_min=self._quant_min,
            observer_algorithm='percentile',
        )

        if operation.type in {'Conv', 'ConvTranspose', 'Gemm'}:
            # set all parameters within Conv, ConvTranspose, Gemm to per-channel quant-config.
            assert operation.num_of_input > 0, 'Seems you got a Conv layer with no parameters.'

            # first parameter must exits, for conv layer it will be conv_weight
            # layout: [out_channel, in_channel, kernel_size, kernel_size]
            if operation.type == 'Conv':
                conv_weight_config = base_quant_config.input_quantization_config[1]
                conv_weight_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_CHANNEL
                )
                conv_weight_config.channel_axis = 0
                conv_weight_config.observer_algorithm = 'minmax'

            elif operation.type == 'ConvTranspose':
                conv_weight_config = base_quant_config.input_quantization_config[1]
                conv_weight_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_CHANNEL
                )
                conv_weight_config.channel_axis = 1
                conv_weight_config.observer_algorithm = 'minmax'

            # first parameter must exits, for gemm layer it will be gemm_weight
            # layout: [in_dim, out_dim]
            elif operation.type == 'Gemm':
                gemm_weight_config = base_quant_config.input_quantization_config[1]
                gemm_weight_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_CHANNEL
                )
                gemm_weight_config.channel_axis = 0
                gemm_weight_config.observer_algorithm = 'minmax'

            # if operation has bias
            # let bias be fp32 because dsp ti don't need quantization parameter of bias
            if operation.num_of_input > 2:
                bias_config = base_quant_config.input_quantization_config[-1]
                bias_config.state = QuantizationStates.FP32

        if operation.type in PASSIVE_OPERATIONS:
            # Those op are not active op.
            base_quant_config.is_active_quant_op = False
        return base_quant_config

    @property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.PPL_DSP_TI_INT8

    @property
    def default_platform(self) -> TargetPlatform:
        return TargetPlatform.FP32

    @property
    def quantize_policy(self) -> QuantizationPolicy:
        return QuantizationPolicy(
            QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_TENSOR
        )

    @property
    def activation_fusion_types(self) -> set:
        return {'Relu', 'Clip'}



from typing import Union

import torch

from esp_ppq.core import (
    PASSIVE_OPERATIONS,
    OperationQuantizationConfig,
    QuantizationPolicy,
    QuantizationProperty,
    QuantizationStates,
    RoundingPolicy,
    TargetPlatform,
)
from esp_ppq.IR import BaseGraph, Operation

from .base import BaseQuantizer


class OnnxruntimeQuantizer(BaseQuantizer):
    def __init__(self, graph: BaseGraph) -> Union[torch.Tensor, list, dict]:
        super().__init__(graph=graph)
        self._num_of_bits = 8
        self._quant_min = -128
        self._quant_max = 127

    def init_quantize_config(self, operation: Operation) -> OperationQuantizationConfig:
        base_quant_config = self.create_default_quant_config(
            policy=self.quantize_policy,
            rounding=self.rounding_policy,
            op=operation,
            num_of_bits=self._num_of_bits,
            exponent_bits=0,
            quant_max=self._quant_max,
            quant_min=self._quant_min,
            observer_algorithm='percentile',
        )

        if operation.type in {'Conv', 'ConvTranspose', 'Gemm'}:
            # set all parameters within Conv, ConvTranspose, Gemm to per-channel quant-config.
            assert operation.num_of_input > 0, 'Seems you got a Conv layer with no parameters.'

            # first parameter must exits, for conv layer it will be conv_weight
            # layout: [out_channel, in_channel, kernel_size, kernel_size]
            if operation.type in {'Conv', 'ConvTranspose'}:
                conv_weight_config = base_quant_config.input_quantization_config[1]
                conv_weight_config._quant_min = -128
                conv_weight_config._quant_max = 127
                conv_weight_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_CHANNEL
                )
                conv_weight_config.channel_axis = 1 if operation.type == 'ConvTranspose' else 0
                conv_weight_config.observer_algorithm = 'minmax'
            # first parameter must exits, for gemm layer it will be gemm_weight
            # layout: [in_dim, out_dim]
            elif operation.type in {'Gemm'}:
                gemm_weight_config = base_quant_config.input_quantization_config[1]
                gemm_weight_config._quant_min = -128
                gemm_weight_config._quant_max = 127
                gemm_weight_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_CHANNEL
                )
                gemm_weight_config.channel_axis = 0
                gemm_weight_config.observer_algorithm = 'minmax'
            # if operation has bias
            if operation.num_of_input > 2:
                bias_config = base_quant_config.input_quantization_config[-1]
                bias_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_CHANNEL
                )
                bias_config.state = QuantizationStates.FP32

        return base_quant_config

    @property
    def quant_operation_types(self) -> set:
        QUANTTYPE = {
            'Conv',
            'ConvTranspose',
            'Gemm',
            'Relu',
            'PRelu',
            'Clip',
            'Pad',
            'Resize',
            'MaxPool',
            'AveragePool',
            'GlobalMaxPool',
            'GlobalAveragePool',
            'Softmax',
            'Mul',
            'Add',
            'Max',
            'Sub',
            'Div',
            'Reshape',
            'LeakyRelu',
            'Concat',
            'Sigmoid',
            'Interp',
            'ReduceMean',
            'Transpose',
            'Slice',
            'Flatten',
            'HardSwish',
            'HardSigmoid',
            'MatMul',
        }
        QUANTTYPE.update(PASSIVE_OPERATIONS)
        return QUANTTYPE

    @property
    def quantize_policy(self) -> QuantizationPolicy:
        return QuantizationPolicy(
            QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_TENSOR
        )

    @property
    def rounding_policy(self) -> RoundingPolicy:
        return RoundingPolicy.ROUND_HALF_EVEN

    @property
    def activation_fusion_types(self) -> set:
        return {'Relu', 'Clip', 'Sigmoid', 'Swish', 'Mish', 'LeakyRelu'}



# Test Quantization System Performance on SuperRes Task(DIV2K)

# Quantizer Configuration
SYMMETRICAL = True
PER_CHANNEL = True
POWER_OF_2 = False
BIT_WIDTH = 8
FP8 = True

# Should contains model file(.onnx)
MODEL_DIR = 'QuantZoo/Model/mmedit'

# Should contains Calib & Test Img Folder
TRAIN_HR_DIR = 'QuantZoo/Data/DIV2K/DIV2K_train_HR'
TRAIN_LR_DIR = 'QuantZoo/Data/DIV2K/DIV2K_train_LR_bicubic'
VALID_HR_DIR = 'QuantZoo/Data/DIV2K/DIV2K_valid_HR'
VALID_LR_DIR = 'QuantZoo/Data/DIV2K/DIV2K_valid_LR_bicubic'

# calibration & test batchsize
# super resolution model must have batchsize = 1
BATCHSIZE = 1

# write report to here
REPORT_DIR = 'QuantZoo/Reports'

CONFIGS = [
    {
        'Model': 'srcnn_x4k915_g1_1000k_div2k',
        'Output': ['output'],
    },
    {
        'Model': 'srgan_x4c64b16_g1_1000k_div2k',
        'Output': ['output'],
    },
    {
        'Model': 'rdn_x4c64b16_g1_1000k_div2k',
        'Output': ['output'],
    },
    {
        'Model': 'edsr_x4c64b16_g1_300k_div2k',
        'Output': ['/generator/conv_last/Conv_output_0'],
    },
]

import os
from typing import Iterable

import torch
from QuantZoo.Data.DIV2K.Data import load_div2k_dataset
from QuantZoo.Data.DIV2K.Eval import psnr, ssim
from QuantZoo.Quantizers import MyFP8Quantizer, MyInt8Quantizer
from QuantZoo.Util import error_analyze, report
from tqdm import tqdm

import esp_ppq.lib as PFL
from esp_ppq import convert_any_to_numpy
from esp_ppq.api import ENABLE_CUDA_KERNEL, load_onnx_graph
from esp_ppq.core import TargetPlatform
from esp_ppq.executor import TorchExecutor
from esp_ppq.IR import BaseGraph
from esp_ppq.quantization.optim import (
    LayerwiseEqualizationPass,
    LearnedStepSizePass,
    ParameterQuantizePass,
    RuntimeCalibrationPass,
)


def evaluation(graph: BaseGraph, dataloader: Iterable, method: str = 'psnr'):
    if method not in {'psnr', 'ssim'}:
        raise Exception('Evaluation method not understood.')
    executor = TorchExecutor(graph)
    ret_collector = []

    for lr_img, hr_img in tqdm(dataloader):
        pred = executor.forward(lr_img.cuda())[0]
        real = hr_img

        # post processing
        pred = convert_any_to_numpy((pred.squeeze(0) * 255).round())
        real = convert_any_to_numpy((real.squeeze(0) * 255).round())

        if method == 'psnr':
            sample_ret = psnr(img1=real, img2=pred, input_order='CHW')
        else:
            sample_ret = ssim(img1=real, img2=pred, input_order='CHW')
        ret_collector.append(sample_ret)

    return sum(ret_collector) / len(ret_collector)


calib_loader = load_div2k_dataset(lr_folder=TRAIN_LR_DIR, hr_folder=TRAIN_HR_DIR)

test_loader = load_div2k_dataset(lr_folder=VALID_LR_DIR, hr_folder=VALID_HR_DIR)

with ENABLE_CUDA_KERNEL():
    for config in CONFIGS:
        model = config['Model']
        monitoring_vars = config['Output']

        print(f"Ready to run quant benchmark on {model}")
        graph = load_onnx_graph(onnx_import_file=os.path.join(MODEL_DIR, model + '.onnx'))

        quantizer = MyInt8Quantizer(
            graph=graph, sym=SYMMETRICAL, per_channel=PER_CHANNEL, power_of_2=POWER_OF_2, num_of_bits=BIT_WIDTH
        )
        if FP8:
            quantizer = MyFP8Quantizer(graph=graph, calibration='floating')

        # convert op to quantable-op
        for name, op in graph.operations.items():
            if op.type in {'Conv', 'ConvTranspose', 'MatMul', 'Gemm'}:
                quantizer.quantize_operation(name, platform=TargetPlatform.INT8)

        # build quant pipeline.
        pipeline = PFL.Pipeline(
            [
                # LayerwiseEqualizationPass(iteration=10),
                ParameterQuantizePass(),
                RuntimeCalibrationPass(),
                # LearnedStepSizePass(steps=500, collecting_device='cpu')
            ]
        )

        # call pipeline.
        executor = TorchExecutor(graph=graph)
        executor.tracing_operation_meta(torch.zeros(size=[BATCHSIZE, 3, 480, 640]).cuda())

        pipeline.optimize(
            graph=graph,
            dataloader=calib_loader,
            verbose=True,
            calib_steps=32,
            collate_fn=lambda x: x[0].cuda(),
            executor=executor,
        )

        result = evaluation(graph=graph, dataloader=test_loader, method='psnr')
        print(f'Model Performance on DIV2K PSNR: {result}')

        # error analyze
        performance = error_analyze(
            graph=graph, outputs=monitoring_vars, dataloader=test_loader, collate_fn=lambda x: x[0].cuda(), verbose=True
        )



import json

import onnx
import torch
from onnx import helper, numpy_helper

from esp_ppq.core import (
    GRAPH_OPSET_ATTRIB,
    ONNX_EXPORT_OPSET,
    ONNX_VERSION,
    PPQ_CONFIG,
    DataType,
    QuantizationStates,
    convert_any_to_numpy,
    ppq_warning,
)
from esp_ppq.core.quant import QuantizationProperty
from esp_ppq.IR import BaseGraph, GraphExporter, Operation, OperationExporter, QuantableOperation, Variable


class ConstantOfShapeExporter(OperationExporter):
    def export(self, operation: Operation, graph: BaseGraph, **kwargs) -> Operation:
        # PATCH 20211203, ConstantOfShape Op causes an export error.
        # 这一问题是由 ConstantOfShape 中的 value 格式问题引发的，下面的代码将导出正确的格式
        operation.attributes["value"] = numpy_helper.from_array(operation.attributes["value"])
        return operation


class MMCVExporter(OperationExporter):
    def export(self, operation: Operation, graph: BaseGraph, **kwargs) -> Operation:
        # MMCV operation must have a domain attribute.
        operation.attributes["domain"] = "mmcv"
        return operation


class InterpExporter(OperationExporter):
    def export(self, operation: Operation, graph: BaseGraph, **kwargs) -> Operation:
        # PATCH 20211216, interp op can not export input_shape attribute.
        operation.attributes.pop("input_shape")
        return operation


class OOSExporter(OperationExporter):
    def export(self, operation: Operation, graph: BaseGraph, **kwargs) -> Operation:
        # MMCV operation must have a domain attribute.
        operation.attributes["domain"] = "com.microsoft"
        return operation


OPERATION_EXPORTERS = {
    "ConstantOfShape": ConstantOfShapeExporter,
    "MMCVRoiAlign": MMCVExporter,
    "grid_sampler": MMCVExporter,
    "Interp": InterpExporter,
    "QAttention": OOSExporter,
    "QGemm": OOSExporter,
    "QLinearAdd": OOSExporter,
    "QLinearAveragePool": OOSExporter,
    "QLinearConcat": OOSExporter,
    "QLinearConv": OOSExporter,
    "QLinearGlobalAveragePool": OOSExporter,
    "QLinearLeakyRelu": OOSExporter,
    "QLinearMul": OOSExporter,
    "QLinearReduceMean": OOSExporter,
    "QLinearSigmoid": OOSExporter,
}


class TengineExporter(GraphExporter):
    def __init__(self) -> None:
        super().__init__()

    def export_quantization_scale(self, scale_path: str, graph: BaseGraph):
        var_scales = {}

        # Render quantization config.
        for operation in graph.operations.values():
            if isinstance(operation, QuantableOperation):
                for config, _var in operation.config_with_variable:
                    if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
                        raise PermissionError('Tengine does not support per channel quantization.')

                    if QuantizationStates.is_activated(config.state) or config.state == QuantizationStates.OVERLAPPED:
                        var_scales[_var.name] = {
                            "scale": config.scale.item(),
                            "zero_point": config.offset.item(),
                        }

        with open(file=scale_path, mode="w") as file:
            for k, v in var_scales.items():
                scale = v["scale"]
                zp = int(round(v["zero_point"]))
                file.write(f"{k} {scale} {zp}\n")

    def export_quantization_config(self, config_path: str, graph: BaseGraph):
        render_buffer = {"configs": {}, "dispatchings": {}, "values": {}}

        # Render quantization config.
        for operation in graph.operations.values():
            if isinstance(operation, QuantableOperation):
                op_dict = {
                    var.name: {
                        "bit_width": config.num_of_bits,
                        "policy": config.policy.to_dict(),
                        "state": config.state.name,
                        "quant_min": config.quant_min,
                        "quant_max": config.quant_max,
                        "hash": config.__hash__(),
                        "dominator": config.dominated_by.__hash__(),
                    }
                    for config, var in operation.config_with_variable
                }

                for config, _ in operation.config_with_variable:
                    if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
                        raise PermissionError('Tengine does not support per channel quantization.')

                    if config.dominated_by == config:
                        render_buffer["values"][config.__hash__()] = {
                            "scale": config.scale.item(),
                            "zero_point": config.offset.item(),
                        }

                render_buffer["configs"][operation.name] = op_dict
                render_buffer["dispatchings"][operation.name] = operation.platform.name

        with open(file=config_path, mode="w") as file:
            json.dump(render_buffer, file, indent=4)

    def export_operation(self, operation: Operation) -> onnx.OperatorProto:
        if operation.type in OPERATION_EXPORTERS:
            exporter = OPERATION_EXPORTERS[operation.type]()
            assert isinstance(exporter, OperationExporter), (
                f"Expected an OpExporter here, however {type(exporter)} was given."
            )
            operation = exporter.export(operation=operation, graph=None)

        attributes = operation.attributes
        for key in attributes:
            value = attributes[key]
            if isinstance(value, DataType):
                attributes[key] = value.value
            if isinstance(value, torch.Tensor):
                attributes[key] = convert_any_to_numpy(value)

        if PPQ_CONFIG.EXPORT_PPQ_INTERNAL_INFO:
            attributes["platform"] = operation.platform.name
        op_proto = helper.make_node(
            op_type=operation.type,
            inputs=[_.name for _ in operation.inputs],
            outputs=[_.name for _ in operation.outputs],
            name=operation.name,
            **attributes,
        )
        return op_proto

    def export_var(self, variable: Variable) -> onnx.TensorProto:
        shape = variable.shape
        dtype = variable.dtype.value

        if dtype is None:
            ppq_warning(
                f"Data type of Variable {variable.name} is not correctly traced, "
                "ppq will export it as fp32 variable to onnx."
            )
            dtype = DataType.FP32.value

        if not variable.is_parameter:
            tensor_proto = helper.make_tensor_value_info(
                name=variable.name,
                # PPQ data type has exact same eunm value with onnx.
                elem_type=dtype,
                shape=shape,
            )
        else:
            value = variable.value
            if isinstance(value, torch.Tensor):
                if value.numel() == 0:
                    value = []
                elif value.ndim >= 1:
                    value = convert_any_to_numpy(variable.value).flatten()
                elif value.ndim == 0:
                    value = [
                        value.item(),
                    ]  # it is fine for onnx, cause shape for this value will be []
            else:
                value = value  # value is python primary type.
            tensor_proto = helper.make_tensor(name=variable.name, data_type=dtype, dims=shape, vals=value)
        return tensor_proto

    def export(self, file_path: str, graph: BaseGraph, config_path: str = None):
        # during export we will remove all boundary operations from graph.
        # we do not want to change the structure of original graph,
        # so there have to take a clone of it.
        # graph = graph.copy()

        name = graph._name
        if not name:
            name = f"{PPQ_CONFIG.NAME} - v({PPQ_CONFIG.VERSION})"

        # if a valid config path is given, export quantization config to there.
        if config_path is not None:
            self.export_quantization_config(config_path, graph)
            scale_path = config_path.replace(".json", ".scale")
            self.export_quantization_scale(scale_path, graph)

        # Ready to export onnx graph defination.
        _inputs, _outputs, _initilizers, _nodes = [], [], [], []
        for operation in graph.topological_sort():
            _nodes.append(self.export_operation(operation))

        for variable in graph.variables.values():
            tensor_proto = self.export_var(variable)
            if variable.name in graph.inputs:
                _inputs.append(tensor_proto)
            if variable.name in graph.outputs:
                _outputs.append(tensor_proto)
            if variable.is_parameter:
                _initilizers.append(tensor_proto)

        graph_def = helper.make_graph(
            name=name,
            nodes=_nodes,
            inputs=_inputs,
            outputs=_outputs,
            initializer=_initilizers,
        )

        # force opset to 11
        if GRAPH_OPSET_ATTRIB not in graph._detail:
            op = onnx.OperatorSetIdProto()
            op.version = ONNX_EXPORT_OPSET
            opsets = [op]
        else:
            opsets = []
            for opset in graph._detail[GRAPH_OPSET_ATTRIB]:
                op = onnx.OperatorSetIdProto()
                op.domain = opset["domain"]
                op.version = opset["version"]
                opsets.append(op)

        onnx_model = helper.make_model(graph_def, producer_name=PPQ_CONFIG.NAME, opset_imports=opsets)
        onnx_model.ir_version = graph._detail.get("ir_version", ONNX_VERSION)
        # onnx.checker.check_model(onnx_model)
        onnx.save(onnx_model, file_path)



from typing import Iterable

import torch
from torch.utils.data import DataLoader

from esp_ppq import BaseGraph, QuantizationSettingFactory, TargetPlatform
from esp_ppq.api import export_ppq_graph, quantize_onnx_model

BATCHSIZE = 32
INPUT_SHAPE = [3, 224, 224]
DEVICE = 'cuda'  # only cuda is fully tested :(  For other executing device there might be bugs.
PLATFORM = TargetPlatform.TRT_INT8  # identify a target platform for your network.
ONNX_PATH = 'Models/cls_model/mobilenet_v2.onnx'


def load_calibration_dataset() -> Iterable:
    return [torch.rand(size=INPUT_SHAPE) for _ in range(32)]


def collate_fn(batch: torch.Tensor) -> torch.Tensor:
    return batch.to(DEVICE)


quant_setting = QuantizationSettingFactory.trt_setting()

# Load training data for creating a calibration dataloader.
calibration_dataset = load_calibration_dataset()
calibration_dataloader = DataLoader(dataset=calibration_dataset, batch_size=BATCHSIZE, shuffle=True)

# quantize your model.
quantized = quantize_onnx_model(
    onnx_import_file=ONNX_PATH,
    calib_dataloader=calibration_dataloader,
    calib_steps=32,
    input_shape=[BATCHSIZE] + INPUT_SHAPE,
    setting=quant_setting,
    collate_fn=collate_fn,
    platform=PLATFORM,
    device=DEVICE,
    verbose=0,
)

# Quantization Result is a PPQ BaseGraph instance.
assert isinstance(quantized, BaseGraph)

# export quantized graph.
export_ppq_graph(
    graph=quantized,
    platform=PLATFORM,
    graph_save_to='Output/quantized(onnx).onnx',
    config_save_to='Output/quantized(onnx).json',
)



# This Example shows you how to replace quantization function with your own quant delegator.

from typing import Iterable

import torch
import torchvision
from torch.utils.data import DataLoader
from tqdm import tqdm

from esp_ppq import *
from esp_ppq.api import quantize_torch_model


class MyQuantDelegator(TorchQuantizeDelegator):
    """Use This class to realize your quantization logic.

    Inherit class TorchQuantizeDelegate, implement interface __call__, then
    register your delegator with executor.register_quantize_delegate
    """

    def __call__(self, tensor: torch.Tensor, config: TensorQuantizationConfig) -> torch.Tensor:
        if config.policy.has_property(QuantizationProperty.ASYMMETRICAL):
            raise ValueError('Sorry, this delegator handles only Symmetrical Quantizations.')
        print('You are invoking cusitmized quant function now.')
        return torch.round(tensor / config.scale) * config.scale


BATCHSIZE = 32
INPUT_SHAPE = [3, 224, 224]
DEVICE = 'cuda'  # only cuda is fully tested :(  For other executing device there might be bugs.
PLATFORM = TargetPlatform.PPL_CUDA_INT8  # identify a target platform for your network.


def load_calibration_dataset() -> Iterable:
    return [torch.rand(size=INPUT_SHAPE) for _ in range(32)]


def collate_fn(batch: torch.Tensor) -> torch.Tensor:
    return batch.to(DEVICE)


# Load a pretrained mobilenet v2 model
model = torchvision.models.mobilenet.mobilenet_v2(pretrained=True)
model = model.to(DEVICE)

# create a setting for quantizing your network with PPL CUDA.
quant_setting = QuantizationSettingFactory.pplcuda_setting()
quant_setting.equalization = True  # use layerwise equalization algorithm.
quant_setting.dispatcher = 'conservative'  # dispatch this network in conservertive way.

# Load training data for creating a calibration dataloader.
calibration_dataset = load_calibration_dataset()
calibration_dataloader = DataLoader(dataset=calibration_dataset, batch_size=BATCHSIZE, shuffle=True)

# quantize your model.
quantized = quantize_torch_model(
    model=model,
    calib_dataloader=calibration_dataloader,
    calib_steps=32,
    input_shape=[BATCHSIZE] + INPUT_SHAPE,
    setting=quant_setting,
    collate_fn=collate_fn,
    platform=PLATFORM,
    onnx_export_file='Output/onnx.model',
    device=DEVICE,
    verbose=0,
)

# Quantization Result is a PPQ BaseGraph instance.
assert isinstance(quantized, BaseGraph)

# build an executor:
executor = TorchExecutor(graph=quantized, device=DEVICE)

# register quant delegator
for var in quantized.variables.values():
    if isinstance(var, QuantableVariable):
        if var.source_op_config is not None:
            executor.register_quantize_delegate(var.source_op_config, MyQuantDelegator)

# run with your network, results are torch.Tensors
for data in tqdm(calibration_dataloader, desc='Running with executor.'):
    results = executor.forward(inputs=data.to(DEVICE))
    break

# remove delegators
for var in quantized.variables.values():
    if isinstance(var, QuantableVariable):
        if var.source_op_config is not None:
            executor.remove_quantize_delegate(var.source_op_config)



class PPQ_GLOBAL_CONFIGURATION:
    def __init__(self) -> None:
        # 是否启动 cuda kernel 加速计算
        self.USING_CUDA_KERNEL = False

        # PPQ 的名字
        self.NAME = 'PPL Quantization Tool'

        # PPQ 的版本号
        self.VERSION = '0.6.6'

        # 导出图时是否导出权重（仅影响 Native 格式导出）
        self.DUMP_VALUE_WHEN_EXPORT = True

        # 导出图时，是否导出调度信息
        self.EXPORT_PPQ_INTERNAL_INFO = False

        # 开启 PPQ 调试模式，将打印所有量化点插入信息
        self.PPQ_DEBUG = False


PPQ_CONFIG = PPQ_GLOBAL_CONFIGURATION()



from typing import Union

import torch

from esp_ppq.core import (
    PASSIVE_OPERATIONS,
    OperationQuantizationConfig,
    QuantizationPolicy,
    QuantizationProperty,
    QuantizationStates,
    RoundingPolicy,
    TargetPlatform,
)
from esp_ppq.IR import BaseGraph, GraphCommandProcessor, Operation

from .base import BaseQuantizer


class TengineQuantizer(BaseQuantizer):
    def __init__(self, graph: Union[BaseGraph, GraphCommandProcessor]) -> Union[torch.Tensor, list, dict]:
        super().__init__(graph=graph)
        self._num_of_bits = 8
        self._quant_min = 0
        self._quant_max = 255

    def init_quantize_config(self, operation: Operation) -> OperationQuantizationConfig:
        base_quant_config = self.create_default_quant_config(
            op=operation,
            num_of_bits=self._num_of_bits,
            exponent_bits=0,
            quant_max=self._quant_max,
            quant_min=self._quant_min,
            observer_algorithm="percentile",
            policy=self.quantize_policy,
            rounding=self.rounding_policy,
        )

        if operation.type in {"Conv", "ConvTranspose", "Gemm"}:
            # if operation has bias
            if operation.num_of_input == 3:
                bias_config = base_quant_config.input_quantization_config[-1]
                # bias should be quantized with 32 bits
                # in python3, int indicates long long in C++
                # so that it has enough precision to represent a number like 2^32
                # however, it may cause a scale underflow
                # here we give bias a 30 bits precision, which is pettery enough in all cases
                bias_config.num_of_bits = 30
                bias_config.quant_max = int(pow(2, 30 - 1) - 1)
                bias_config.quant_min = -int(pow(2, 30 - 1))
                bias_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_TENSOR
                )
                bias_config.state = QuantizationStates.PASSIVE_INIT
            for config in base_quant_config.input_quantization_config[1:]:
                config.observer_algorithm = "minmax"

        if operation.type in PASSIVE_OPERATIONS:
            # Those op are not active op.
            base_quant_config.is_active_quant_op = False
        return base_quant_config

    @property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.TENGINE_INT8

    @property
    def default_platform(self) -> TargetPlatform:
        return TargetPlatform.FP32

    @property
    def quant_operation_types(self) -> set:
        return {
            "Conv",
            "GlobalAveragePool",
            "AveragePool",
            "Relu",
            "Add",
            "Mul",
            "Clip",
            "Sigmoid",
            "MatMul",
            "Gemm",
            "Concat",
            "LeakyRelu",
            "Softmax",
        }

    @property
    def quantize_policy(self) -> QuantizationPolicy:
        return QuantizationPolicy(
            QuantizationProperty.ASYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_TENSOR
        )

    @property
    def rounding_policy(self) -> RoundingPolicy:
        return RoundingPolicy.ROUND_HALF_EVEN

    @property
    def activation_fusion_types(self) -> set:
        return {"Relu", "Clip"}



from typing import List

import torch

from esp_ppq.IR import Operation

from .base import ASSERT_NUM_OF_INPUT, TorchBackendContext
from .default import DEFAULT_BACKEND_TABLE

PPL_GPU_BACKEND_TABLE = DEFAULT_BACKEND_TABLE.copy()


# When you trying to implement a custimized function for ppl_gpu platform
# Be aware that you can just overwrite part of DEFAULT_DISPATCHING_TABLE
# rather than rewrite all dispatching table.
# here an example was given: Sample_Forward
def Sample_Forward():
    return None


def AveragePool_forward(
    op: Operation, values: List[torch.Tensor], ctx: TorchBackendContext = None, **kwargs
) -> torch.Tensor:
    """AveragePool-11 AveragePool consumes an input tensor X and applies
    average pooling across the tensor according to kernel sizes, stride sizes,
    and pad lengths. average pooling consisting of computing the average on all
    values of a subset of the input tensor according to the kernel size and
    downsampling the data into the output tensor Y for further processing. The
    output spatial shape will be following:

    Attributes
        auto_pad : string (default is NOTSET)
        auto_pad must be either NOTSET, SAME_UPPER, SAME_LOWER or VALID.

        Where default value is NOTSET, which means explicit padding is used.
        SAME_UPPER or SAME_LOWER mean pad the input so that `output_shape[i] = ceil(input_shape[i] / strides[i])`
            for each axis `i`. The padding is split between the two sides equally or almost equally
            (depending on whether it is even or odd).

        In case the padding is an odd number,
            the extra padding is added at the end for SAME_UPPER and at the beginning for SAME_LOWER.

    ceil_mode : int (default is 0)
        Whether to use ceil or floor (default) to compute the output shape.

    count_include_pad : int (default is 0)
        Whether include pad pixels when calculating values for the edges. Default is 0, doesn't count include pad.

    kernel_shape : list of ints (required)
        The size of the kernel along each axis.

    pads : list of ints
        Padding for the beginning and ending along each spatial axis,
        it can take any value greater than or equal to 0.

        The value represent the number of pixels added to the beginning and end part of the corresponding axis.
        `pads` format should be as follow [x1_begin, x2_begin...x1_end, x2_end,...],
            where xi_begin the number of pixels added at the beginning of axis `i` and xi_end,
            the number of pixels added at the end of axis `i`.

        This attribute cannot be used simultaneously with auto_pad attribute.
        If not present, the padding defaults to 0 along start and end of each spatial axis.

    strides : list of ints
        Stride along each spatial axis.
        If not present, the stride defaults to 1 along each spatial axis.

    Inputs
        X (differentiable) : T
        Input data tensor from the previous operator;
            dimensions for image case are (N x C x H x W),
            where N is the batch size, C is the number of channels,
            and H and W are the height and the width of the data.

        For non image case, the dimensions are in the form of (N x C x D1 x D2 ... Dn),
            where N is the batch size. Optionally, if dimension denotation is in effect,
            the operation expects the input data tensor to arrive with
            the dimension denotation of [DATA_BATCH, DATA_CHANNEL, DATA_FEATURE, DATA_FEATURE ...].

    Outputs
        Y (differentiable) : T
        Output data tensor from average or max pooling across the input tensor.
            Dimensions will vary based on various kernel, stride, and pad sizes.
            Floor value of the dimension is used

    Args:
        op (Operation): [description]
        values (List[torch.Tensor]): [description]
        ctx (TorchBackendContext, optional): [description]. Defaults to None.

    Returns:
        torch.Tensor: [description]
    """
    ASSERT_NUM_OF_INPUT(op=op, values=values, min_num_of_input=1, max_num_of_input=1)
    # checker(op.attributes, input_value[0].shape[2:])
    # op_attr = preprocess_attr(op.attributes, 'Pooling')
    # [input_value] = input_value
    # if op.type == 'GlobalAveragePool':
    #     image_shape = input_value.size()[-2:]
    #     op_attr['kernel_size'] = image_shape
    #
    # output = F.avg_pool2d(input_value, **op_attr)
    # return output


PPL_GPU_BACKEND_TABLE['Sample_Function'] = Sample_Forward



from esp_ppq.core import NetworkFramework, TargetPlatform
from esp_ppq.executor import register_operation_handler
from esp_ppq.IR import GraphBuilder, GraphExporter
from esp_ppq.quantization.observer import OBSERVER_TABLE, OperationObserver
from esp_ppq.quantization.quantizer import BaseQuantizer

from .common import __EXPORTERS__, __PARSERS__, __QUANTIZER_COLLECTION__


def register_network_quantizer(quantizer: type, platform: TargetPlatform):
    """Register a quantizer to ppq quantizer collection.

    This function will override the default quantizer collection:
        register_network_quantizer(MyQuantizer, TargetPlatform.TRT_INT8) will replace the default TRT_INT8 quantizer.

    Quantizer should be a subclass of BaseQuantizer, do not provide an instance here as ppq will initilize it later.
    Your quantizer must require no initializing params.

    Args:
        quantizer (type): quantizer to be inserted.
        platform (TargetPlatform): corresponding platfrom of your quantizer.
    """
    if not isinstance(quantizer, type):
        raise TypeError(
            f'You can only register a class type as custimized ppq quantizer, '
            f'however {type(quantizer)} is given. '
            '(Requiring a class type here, do not provide an instance)'
        )
    if not issubclass(quantizer, BaseQuantizer):
        raise TypeError('You can only register a subclass of BaseQuantizer as custimized quantizer.')
    __QUANTIZER_COLLECTION__[platform] = quantizer


def register_network_parser(parser: type, framework: NetworkFramework):
    """Register a parser to ppq parser collection.

    This function will override the default parser collection:
        register_network_parser(MyParser, NetworkFramework.ONNX) will replace the default ONNX parser.

    Parser should be a subclass of GraphBuilder, do not provide an instance here as ppq will initilize it later.
    Your quantizer must require no initializing params.

    Args:
        parser (type): parser to be inserted.
        framework (NetworkFramework): corresponding NetworkFramework of your parser.
    """
    if not isinstance(parser, type):
        raise TypeError(
            f'You can only register a class type as custimized ppq parser, '
            f'however {type(parser)} is given. '
            f'(Requiring a class type here, do not provide an instance)'
        )
    if not issubclass(parser, GraphBuilder):
        raise TypeError('You can only register a subclass of GraphBuilder as custimized parser.')
    __PARSERS__[framework] = parser


def register_network_exporter(exporter: type, platform: TargetPlatform):
    """Register an exporter to ppq exporter collection.

    This function will override the default exporter collection:
        register_network_quantizer(MyExporter, TargetPlatform.TRT_INT8) will replace the default TRT_INT8 exporter.

    Exporter should be a subclass of GraphExporter, do not provide an instance here as ppq will initilize it later.
    Your Exporter must require no initializing params.

    Args:
        exporter (type): exporter to be inserted.
        platform (TargetPlatform): corresponding platfrom of your exporter.
    """
    if not isinstance(exporter, type):
        raise TypeError(
            f'You can only register a class type as custimized ppq exporter, '
            f'however {type(exporter)} is given. '
            f'(Requiring a class type here, do not provide an instance)'
        )
    if not issubclass(exporter, GraphExporter):
        raise TypeError('You can only register a subclass of GraphExporter as custimized exporter.')
    __EXPORTERS__[platform] = exporter


def register_calibration_observer(algorithm: str, observer: type):
    """Register an calibration observer to OBSERVER_TABLE.

    This function will override the existing OBSERVER_TABLE without warning.

    registed observer must be a sub class of OperationObserver.

    Args:
        exporter (type): exporter to be inserted.
        platform (TargetPlatform): corresponding platfrom of your exporter.
    """
    if not isinstance(observer, type):
        raise TypeError(
            f'You can only register an OperationObserver as custimized ppq observer, '
            f'however {type(observer)} is given. '
        )
    if not issubclass(observer, OperationObserver):
        raise TypeError('Regitsing observer must be a subclass of OperationObserver.')
    OBSERVER_TABLE[algorithm] = observer


__all__ = [
    'register_network_quantizer',
    'register_network_parser',
    'register_network_exporter',
    'register_operation_handler',
    'register_calibration_observer',
]



# -*- coding: utf-8 -*-
import os
from functools import partial
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Tuple,
    Union,
)

import numpy as np
import onnx
import torch
from onnxsim import simplify
from torch.utils.data import DataLoader

import esp_ppq.lib as PFL
from esp_ppq.api.interface import load_onnx_graph, quantize_onnx_model
from esp_ppq.api.setting import QuantizationSetting, QuantizationSettingFactory
from esp_ppq.core import QuantizationVisibility, TargetPlatform, empty_ppq_cache
from esp_ppq.executor import BaseGraphExecutor, TorchExecutor
from esp_ppq.IR import BaseGraph
from esp_ppq.log import NaiveLogger
from esp_ppq.quantization.analyse import graphwise_error_analyse
from esp_ppq.quantization.analyse.layerwise import layerwise_error_analyse
from esp_ppq.quantization.optim import *

logger = NaiveLogger.get_logger('ESPDL')


def get_target_platform(
    target: str,
    num_of_bits: int = 8,
    float: bool = False,
    **kwargs: Any,
):
    """Quantize onnx model and return quantized ppq graph and executor .

    Args:
        hi_precision (bool, optional): When the operator is quantified at 16-bit, does PPQ perform forward calculations with double precision,
                                    which, although maintaining high precision at 16-bit, may differ in precision from ESP-DL's operator.
                                    Currently, conv2d, gemm, and matmal are consistent in precision with PPQ's quantized results when calculated
                                    with double precision during 16-bit operations.
    """

    platform = None
    target = target.lower()
    hi_precision = kwargs.get('hi_precision', False)

    if float:
        platform = TargetPlatform.FP32
    else:
        if num_of_bits == 8 and target == "esp32p4":
            platform = TargetPlatform.ESPDL_INT8
        elif num_of_bits == 16 and target == "esp32p4" and not hi_precision:
            platform = TargetPlatform.ESPDL_INT16
        elif num_of_bits == 8 and target == "esp32s3":
            platform = TargetPlatform.ESPDL_S3_INT8
        elif num_of_bits == 16 and target == "esp32s3" and not hi_precision:
            platform = TargetPlatform.ESPDL_S3_INT16
        elif num_of_bits == 16 and target == "esp32p4" and hi_precision:
            platform = TargetPlatform.ESPDL_H_PRE_INT16
        elif num_of_bits == 16 and target == "esp32s3" and hi_precision:
            platform = TargetPlatform.ESPDL_S3_H_PRE_INT16
        elif num_of_bits == 8 and target == "c":
            platform = TargetPlatform.ESPDL_C_INT8
        elif num_of_bits == 16 and target == "c" and not hi_precision:
            platform = TargetPlatform.ESPDL_C_INT16
        elif num_of_bits == 16 and target == "c" and hi_precision:
            platform = TargetPlatform.ESPDL_C_H_PRE_INT16
        else:
            platform = TargetPlatform.FP32
            logger.warning(f"num_of_bits:{num_of_bits} and target:{target} will return TargetPlatform.FP32")

    return platform


def insert_streaming_cache_on_var(
    var_name: str, window_size: int, op_name: str = None, frame_axis: int = 1
) -> Dict[str, Any]:
    """
    Get streaming attributes dictionary.
    Args:
        var_name (str): Name of the variable.
        window_size (int): Size of the streaming window.
        op_name (str, optional): Name of the src or dest operation. Defaults to None.
        frame_axis (int, optional): Axis representing the frame dimension. Defaults to 1. (NCH/NCHW: 2, NHWC: 1)
    Returns:
        Dict[str, Any]: Dictionary containing streaming attributes.
    """
    attributes = {
        'window_size': window_size,
        'op_name': op_name,
        'frame_axis': frame_axis,
    }
    return {var_name: attributes}


def get_random_inputs(input_shape: List[Any], dtype=torch.float32, device='cpu') -> List[Any]:
    if not isinstance(input_shape[0], list):
        input_shape = [input_shape]

    inputs = [torch.randn(size=shape, device=device, dtype=dtype) for shape in input_shape]

    return inputs


def generate_test_value(
    graph: BaseGraph,
    running_device: str,
    inputs: Union[dict, list, torch.Tensor],
    output_names: List[str] = None,
) -> Dict[str, Dict[str, np.ndarray]]:
    test_inputs_value = {}
    test_outputs_value = {}

    executor = TorchExecutor(graph=graph, device=running_device)
    outputs = executor.forward(inputs=inputs, output_names=output_names)
    # get test_inputs_value
    if isinstance(inputs, dict):
        for name, value in inputs.items():
            if name in graph.inputs:
                test_inputs_value[name] = value.clone().detach().to(running_device)
            else:
                logger.error(f"Can not find input {name} in your graph inputs, please check.")
    else:
        inputs_tmp = executor.prepare_input(inputs=inputs)
        test_inputs_value = {name: value.clone().detach().to(running_device) for name, value in inputs_tmp.items()}

    # get test_outputs_value
    if output_names is None:
        outputs_dictionary = graph.outputs
        test_outputs_value = {
            key: outputs[idx].clone().detach().to(running_device) for idx, key in enumerate(outputs_dictionary)
        }
    else:
        test_outputs_value = {
            output_name: output.clone().detach().to(running_device)
            for output_name, output in zip(output_names, outputs)
        }

    return {"inputs": test_inputs_value, "outputs": test_outputs_value}


def collate_fn_template(batch: Union[torch.Tensor, List[torch.Tensor]], dtype=torch.float32, device='cpu'):
    if isinstance(batch, list) and isinstance(batch[0], torch.Tensor):
        return [x.type(dtype).to(device) for x in batch]
    elif isinstance(batch, torch.Tensor):
        return batch.type(dtype).to(device)
    else:
        logger.error("please provide a valid collate_fn.")


@empty_ppq_cache
def espdl_quantize_onnx(
    onnx_import_file: str,
    espdl_export_file: str,
    calib_dataloader: DataLoader,
    calib_steps: int,
    input_shape: List[Any],
    inputs: List[Any] = None,
    target: str = "esp32p4",
    num_of_bits: int = 8,
    collate_fn: Callable = None,
    dispatching_override: Dict[str, TargetPlatform] = None,
    dispatching_method: str = "conservative",
    setting: QuantizationSetting = None,
    device: str = "cpu",
    error_report: bool = True,
    skip_export: bool = False,
    export_config: bool = True,
    export_test_values: bool = False,
    test_output_names: List[str] = None,
    auto_streaming: bool = False,
    streaming_table: Dict[str, Dict[str, Any]] = None,
    streaming_input_shape: List[Any] = None,
    verbose: int = 0,
    **kwargs: Any,
) -> BaseGraph:
    """Quantize onnx model and return quantized ppq graph and executor .

    Args:
        onnx_import_file (str): onnx model file path
        calib_dataloader (DataLoader): calibration data loader
        calib_steps (int): calibration steps
        input_shape (List[int]):a list of ints indicating size of inputs and batch size must be 1
        inputs (List[str]): a list of Tensor and batch size must be 1
        target: target chip, support "esp32p4", "esp32s3" or "c"
        num_of_bits: the number of quantizer bits, 8 or 16
        collate_fn (Callable): batch collate func for preprocessing
        dispatching_override (deprecated): override dispatching result.
                                           It is recommended to use the `setting` parameter, as this argument will be removed in the next version.
        dispatching_method (deprecated): Refer to https://github.com/espressif/esp-ppq/blob/master/ppq/scheduler/__init__.py#L8.
                                         It is recommended to use the `setting` parameter, as this argument will be removed in the next version.
        setting (QuantizationSetting): Quantization setting, default espdl setting will be used when set None
        device (str, optional):  execution device, defaults to 'cpu'.
        error_report (bool, optional): whether to print error report, defaults to True.
        skip_export (bool, optional): whether to export the quantized model, defaults to False.
        export_config (bool, optional): whether to export the quantization configuration, defaults to True.
        export_test_values (bool, optional): whether to export the test values, defaults to False.
        test_output_names (List[str], optional): tensor names of the model want to test, defaults to None.
        verbose (int, optional): whether to print details, defaults to 0.
        hi_precision (bool, optional): It's in kwargs. When the operator is quantified at 16-bit, does PPQ perform forward calculations with double precision,
                                       which, although maintaining high precision at 16-bit, may differ in precision from ESP-DL's operator.
                                       Currently, conv2d, gemm, and matmal are consistent in precision with PPQ's quantized results when calculated
                                       with double precision during 16-bit operations.
        metadata_props (Dict[str, str], optional): It's in kwargs. You can add custom key-value pairs to the model.

    Returns:
        BaseGraph:      The Quantized Graph, containing all information needed for backend execution
    """

    model = onnx.load(onnx_import_file)
    model_sim, check = simplify(model)
    if check:
        onnx.save(model_sim, onnx_import_file)

    export_path = os.path.dirname(os.path.abspath(espdl_export_file))
    os.makedirs(export_path, exist_ok=True)

    # ------------------------------------------------------------
    #
    # 1: Quantize ONNX Model.
    #
    #  ------------------------------------------------------------
    if calib_dataloader is None or calib_steps is None:
        raise TypeError("Quantization needs a valid calib_dataloader and calib_steps setting.")
    target_platform = get_target_platform(target, num_of_bits, **kwargs)
    input_dtype = torch.float32

    if not collate_fn:
        collate_fn = partial(collate_fn_template, dtype=input_dtype, device=device)

    ppq_graph = None
    if inputs:
        dummy_inputs = inputs
    else:
        dummy_inputs = get_random_inputs(input_shape, input_dtype, device)

    if target_platform != TargetPlatform.FP32:
        if dispatching_override is not None or dispatching_method != "conservative":
            logger.warning(
                "It is recommended to use the setting parameter. The dispatching_override and dispatching_method will be deprecated."
            )

        if setting is None:
            setting = QuantizationSettingFactory.espdl_setting()

        if dispatching_method != "conservative":
            setting.dispatcher = dispatching_method

        if dispatching_override is not None:
            for opname, platform in dispatching_override.items():
                setting.dispatching_table.append(opname, platform)

        ppq_graph = quantize_onnx_model(
            onnx_import_file=onnx_import_file,
            calib_dataloader=calib_dataloader,
            calib_steps=calib_steps,
            input_shape=None,
            platform=target_platform,
            input_dtype=input_dtype,
            inputs=dummy_inputs,
            setting=setting,
            collate_fn=collate_fn,
            device=device,
            verbose=verbose,
            do_quantize=True,
        )

        # ------------------------------------------------------------
        #
        # 2: Analyze Quantization Errors.
        #
        # ------------------------------------------------------------
        if error_report:
            graphwise_error_analyse(
                graph=ppq_graph,
                running_device=device,
                collate_fn=collate_fn,
                dataloader=calib_dataloader,
            )

            layerwise_error_analyse(
                graph=ppq_graph,
                running_device=device,
                collate_fn=collate_fn,
                dataloader=calib_dataloader,
            )
    else:
        # support TargetPlatform.FP32
        ppq_graph = load_onnx_graph(onnx_import_file=onnx_import_file)
        executor = TorchExecutor(graph=ppq_graph, device=device)
        executor.tracing_operation_meta(inputs=dummy_inputs)
        target_platform = TargetPlatform.ESPDL_INT8

    # ------------------------------------------------------------
    #
    # 3: Export ESPDL Model.
    #
    # ------------------------------------------------------------
    if not skip_export:
        values_for_test = None
        if export_test_values:
            values_for_test = generate_test_value(ppq_graph, device, dummy_inputs, test_output_names)

        PFL.Exporter(platform=target_platform).export(
            file_path=espdl_export_file,
            graph=ppq_graph,
            values_for_test=values_for_test,
            export_config=export_config,
            auto_streaming=auto_streaming,
            streaming_table=streaming_table,
            streaming_input_shape=streaming_input_shape,
            **kwargs,
        )
    return ppq_graph


def espdl_quantize_torch(
    model: torch.nn.Module,
    espdl_export_file: str,
    calib_dataloader: DataLoader,
    calib_steps: int,
    input_shape: List[Any],
    inputs: List[Any] = None,
    target: str = "esp32p4",
    num_of_bits: int = 8,
    collate_fn: Callable = None,
    dispatching_override: Dict[str, TargetPlatform] = None,
    dispatching_method: str = "conservative",
    setting: QuantizationSetting = None,
    device: str = "cpu",
    error_report: bool = True,
    skip_export: bool = False,
    export_config: bool = True,
    export_test_values: bool = False,
    test_output_names: List[str] = None,
    verbose: int = 0,
    opset_version: int = 18,
    auto_streaming: bool = False,
    streaming_table: Dict[str, Dict[str, Any]] = None,
    streaming_input_shape: Dict[str, List[int]] = None,
    **kwargs: Any,
) -> BaseGraph:
    """Quantize torch model and return quantized ppq graph and executor .

    Args:
        model (torch.nn.Module): torch model
        calib_dataloader (DataLoader): calibration data loader
        calib_steps (int): calibration steps
        input_shape (List[int]):a list of ints indicating size of inputs and batch size must be 1
        inputs (List[str]): a list of Tensor and batch size must be 1
        target: target chip, support "esp32p4", "esp32s3" or "c"
        num_of_bits: the number of quantizer bits, 8 or 16
        collate_fn (Callable): batch collate func for preprocessing
        dispatching_override (deprecated): override dispatching result.
                                        It is recommended to use the `setting` parameter, as this argument will be removed in the next version.
        dispatching_method (deprecated): Refer to https://github.com/espressif/esp-ppq/blob/master/ppq/scheduler/__init__.py#L8.
                                        It is recommended to use the `setting` parameter, as this argument will be removed in the next version.
        setting (QuantizationSetting): Quantization setting, default espdl setting will be used when set None
        device (str, optional):  execution device, defaults to 'cpu'.
        error_report (bool, optional): whether to print error report, defaults to True.
        skip_export (bool, optional): whether to export the quantized model, defaults to False.
        export_config (bool, optional): whether to export the quantization configuration, defaults to True.
        export_test_values (bool, optional): whether to export the test values, defaults to False.
        test_output_names (List[str], optional): tensor names of the model want to test, defaults to None.
        verbose (int, optional): whether to print details, defaults to 0.
        hi_precision (bool, optional): It's in kwargs. When the operator is quantified at 16-bit, does PPQ perform forward calculations with double precision,
                                       which, although maintaining high precision at 16-bit, may differ in precision from ESP-DL's operator.
                                       Currently, conv2d, gemm, and matmal are consistent in precision with PPQ's quantized results when calculated
                                       with double precision during 16-bit operations.
        metadata_props (Dict[str, str], optional): It's in kwargs. You can add custom key-value pairs to the model.

    Returns:
        BaseGraph:      The Quantized Graph, containing all information needed for backend execution
    """
    if not isinstance(input_shape[0], list):
        input_shape = [input_shape]
    export_path = os.path.dirname(os.path.abspath(espdl_export_file))
    os.makedirs(export_path, exist_ok=True)

    # step1: export onnx model
    model = model.eval()
    model = model.to(device)

    base_file_name, _ = os.path.splitext(espdl_export_file)
    onnx_file_path = base_file_name + ".onnx"
    if torch.__version__ >= "2.9.0":
        torch.onnx.export(
            model=model,
            args=tuple(
                [
                    torch.zeros(
                        size=shape,
                        device=device,
                        dtype=torch.float32,
                    )
                    for shape in input_shape
                ]
            ),
            f=onnx_file_path,
            opset_version=opset_version,
            do_constant_folding=True,
            dynamo=False,
        )
    else:
        torch.onnx.export(
            model=model,
            args=tuple(
                [
                    torch.zeros(
                        size=shape,
                        device=device,
                        dtype=torch.float32,
                    )
                    for shape in input_shape
                ]
            ),
            f=onnx_file_path,
            opset_version=opset_version,
            do_constant_folding=True,
        )

    # step2: quantize onnx model and export espdl model
    return espdl_quantize_onnx(
        onnx_import_file=onnx_file_path,
        espdl_export_file=espdl_export_file,
        calib_dataloader=calib_dataloader,
        calib_steps=calib_steps,
        input_shape=input_shape,
        inputs=inputs,
        target=target,
        num_of_bits=num_of_bits,
        collate_fn=collate_fn,
        dispatching_override=dispatching_override,
        dispatching_method=dispatching_method,
        setting=setting,
        device=device,
        error_report=error_report,
        skip_export=skip_export,
        export_config=export_config,
        export_test_values=export_test_values,
        test_output_names=test_output_names,
        auto_streaming=auto_streaming,
        streaming_table=streaming_table,
        streaming_input_shape=streaming_input_shape,
        verbose=verbose,
        **kwargs,
    )



"""PPQ Core Data Structure Abstraction PPQ 核心量化结构抽象.

You are not allowed to modify this 请勿修改此文件
"""

import time  # for hash generation
from enum import Enum
from typing import Any, Iterable, List

import torch

from .common import EXPORT_OVERLAPPED_CONFIG
from .defs import ppq_warning
from .storage import Serializable

MAX_RECURSION_DEPTH = 10000
import sys

sys.setrecursionlimit(MAX_RECURSION_DEPTH)


class QuantizationVisibility(Enum):
    FORCE_EXPORT = 1
    EXPORT_WHEN_ACTIVE = 2
    INTERNAL = 3


class NetworkFramework(Enum):
    PPL = 1
    ONNX = 2
    CAFFE = 3
    NXP = 4
    NATIVE = 5


class TargetPlatform(Enum):
    """TargetPlatform is a core abstraction of PPQ framework, it defines
    "platform" as an attribute of an operation. Platform attribute of an
    operation indicates where this operation is going to be deployed. This
    feature enables PPQ to simulate inter-device computing.

    Platform attribute also tells PPQ how to quantize an operation, and how to execute it.
        ATTENTION: Different platform might bring different behaviour of a same operation.
        ATTENTION: Operation which is assigned to an non-quantizible platform will never be quantized.

    There are several supported platforms for PPQ now,
        however you are supposed to be aware of some particular platforms here:

    SHAPE_OR_INDEX is a virtual platform, however it is an EXTREMELY IMPORTANT components in esp_ppq.
        Dispatch an operation to platform SHAPE_OR_INDEX means this operation is SOI-related,
        it processes a SOI tensor and gives a processed SOI, all calculation of this operation must be sent to CPU
            (or any platform capable for calculating this.) when deploy.

        An operation with SHAPE_OR_INDEX platform assigned will never be quantized regardless of its type.
        It is a crucial feature for quantizing network that contains SOI-related operation. (Shufflenet etc.)

        By default, PPQ automatically detects all SOI-related operations, and dispatch them to SHAPE_OR_INDEX platform.
        To understand how this feature works, see also: esp_ppq.sche

    UNSPECIFIED is a virtual platform, all operations are sent to this platform once they were created.
        Quantizer then dispatches them towards desired platform through its quantization logic.
    """

    MNN_INT8 = 100
    TRT_INT8 = 101
    TRT_FP8 = 105
    NCNN_INT8 = 102
    OPENVINO_INT8 = 103
    TENGINE_INT8 = 104
    ASC_INT8 = 106

    PPL_CUDA_INT8 = 201
    PPL_CUDA_INT4 = 202
    PPL_CUDA_FP16 = 203
    PPL_CUDA_MIX = 204

    PPL_DSP_INT8 = 301
    SNPE_INT8 = 302
    PPL_DSP_TI_INT8 = 303
    QNN_DSP_INT8 = 304

    HOST_INT8 = 401

    NXP_INT8 = 501
    FPGA_INT8 = 502

    ESPDL_INT8 = 551
    ESPDL_INT16 = 552
    ESPDL_S3_INT8 = 553
    ESPDL_S3_INT16 = 554
    ESPDL_H_PRE_INT16 = 555
    ESPDL_S3_H_PRE_INT16 = 556
    ESPDL_C_INT8 = 557
    ESPDL_C_INT16 = 558
    ESPDL_C_H_PRE_INT16 = 559

    RKNN_INT8 = 601

    METAX_INT8_C = 701  # channel wise
    METAX_INT8_T = 702  # tensor wise

    HEXAGON_INT8 = 801
    GRAPHCORE_FP8 = 901

    FP32 = 0
    FP16 = 1
    BF16 = 2
    FP8 = 3
    INT8 = 4
    # SHAPE-OR-INDEX
    SOI = -1
    # initial state
    UNSPECIFIED = -2
    # boundary op
    BOUNDARY = -3
    # just used for calling exporter
    ONNX = -4
    NATIVE = -6
    ONNXRUNTIME = -7
    # THIS IS A DUUMY PLATFORM JUST FOR CREATING YOUR OWN EXTENSION.
    EXTENSION = -10086

    @classmethod
    def is_quantized_platform(cls, platform) -> bool:
        # removed since PPQ 0.6.6
        return platform in {
            cls.PPL_DSP_INT8,
            cls.PPL_DSP_TI_INT8,
            cls.QNN_DSP_INT8,
            cls.TRT_INT8,
            cls.NCNN_INT8,
            cls.NXP_INT8,
            cls.SNPE_INT8,
            cls.PPL_CUDA_INT8,
            cls.PPL_CUDA_INT4,
            cls.EXTENSION,
            cls.PPL_CUDA_MIX,
            cls.RKNN_INT8,
            cls.METAX_INT8_C,
            cls.METAX_INT8_T,
            cls.OPENVINO_INT8,
            cls.FPGA_INT8,
            cls.TENGINE_INT8,
            cls.FP8,
            cls.GRAPHCORE_FP8,
            cls.TRT_FP8,
            cls.ASC_INT8,
            cls.UNSPECIFIED,
            cls.INT8,
            cls.MNN_INT8,
        }


class RoundingPolicy(Enum):
    """RoundingPolicy is a core setting for PPQ quantization calculation. It
    defines rounding behaviour inside quantization calculation.

    Formula: quant(x) = clip(round(x / scale, RoundingPolicy), -128, 127)

    PPQ Supports 7 different rounding policies now.
    Take a look at https://en.wikipedia.org/wiki/Rounding

    ATTENTION: RoundingPolicy greatly affects PPQ executor behaviour in some cases,
        to get a correct result from PPQ executor,
        make sure your RoundingPolicy is the same as your hardware.
    """

    ROUND_HALF_EVEN = 0
    ROUND_HALF_UP = 1
    ROUND_HALF_DOWN = 2
    ROUND_HALF_TOWARDS_ZERO = 3
    ROUND_HALF_FAR_FORM_ZERO = 4
    ROUND_TO_NEAR_INT = 5
    ROUND_UP = 6
    ROUND_DOWN = 7


class QuantizationProperty(Enum):
    """QuantizationProperty is a core abstraction for PPQ quantization
    calculation. QuantizationProperty and QuantizationPolicy together build a
    bitmap to describe quantization policy.

    A QuantizationPolicy instance contains multiple QuantizationProperty,
        QuantizationPolicy is used in PPQ (alone with other configuration) to describe how a tensor is quantized.

    During simulating, executor will quantize tensor corresponding to its QuantizationPolicy.
        (QuantizationPolicy is included by TensorQuantizationConfig)

    There are 8 different quantization property(s) supported by PPQ now.

        PER_TENSOR: Also known as per-layer quantization, which mean all parameters of this layer share the same scale and offset.
            (For Convulution layer and Gemm layer which has bias, bias layer will be negative quantized, they do not have a valid scale)

        PER_CHANNEL: parameters are quantized alone channel axis, each channel has a stand-alone scale and offset.

        LINEAR: Linear quantization, follow formula: quant(x) = clip(round(x / scale))

        FLOATING: Low precision float quantization, FP8, BF16, FP16.

        SYMMETRICAL: Symmetrical quantization, offset is deactivated in this mode.

        ASYMMETRICAL: Asymmetrical quantization, offset is activated in this mode.

        POWER_OF_2: Power-of-2 quantization, scale must be pow(2, k) in this mode.

        DYNAMIC: Dynamic Activation Quantization, scale is computed on the fly.

    ATTENTION: Not all combinations of all 8 QuantizationProperty are valid, see QuantizationPolicy.__check_valid
    ATTENTION: QuantizationPolicy is read-only, user can only assign its value when created, the only interface of
        QuantizationPolicy is function QuantizationPolicy.has_property.
    """

    PER_TENSOR = 0x00000001
    PER_CHANNEL = 0x00000002
    LINEAR = 0x00000004
    FLOATING = 0x00000008
    SYMMETRICAL = 0x00000010
    ASYMMETRICAL = 0x00000020
    POWER_OF_2 = 0x00000040
    DYNAMIC = 0x00000080

    def __or__(self, other: int) -> int:
        return self.value + other

    def __ror__(self, other: int) -> int:
        return self.value + other

    def __and__(self, other: int) -> int:
        return self.value & other

    def __rand__(self, other: int) -> int:
        return self.value & other

    def __radd__(self, other: int) -> int:
        return self.value + other

    def __add__(self, other: int) -> int:
        return self.value + other

    def __sub__(self, other: int) -> int:
        return self - (self.value & other)

    def __rsub__(self, other: int) -> int:
        return other - (self.value & other)


class QuantizationPolicy:
    """QuantizationPolicy is a core abstraction for PPQ quantization
    calculation. QuantizationProperty and QuantizationPolicy together build a
    bitmap to describe quantization policy.

    A QuantizationPolicy instance contains multiple QuantizationProperty,
        QuantizationPolicy is used in PPQ (alone with other configuration) to describe how a tensor is quantized.

    During simulating, executor will quantize tensor corresponding to its QuantizationPolicy.
        (QuantizationPolicy is included by TensorQuantizationConfig)

    There are 8 different quantization property(s) supported by PPQ now.

        PER_TENSOR: Also known as per-layer quantization, which mean all parameters of this layer share the same scale and offset.
            (For Convulution layer and Gemm layer which has bias, bias layer will be negative quantized, they do not have a valid scale)

        PER_CHANNEL: parameters are quantized alone channel axis, each channel has a stand-alone scale and offset.

        LINEAR: Linear quantization, follow formula: quant(x) = clip(round(x / scale))

        EXPONENTIAL: Exponential quantization, not yet used.

        SYMMETRICAL: Symmetrical quantization, offset is deactivated in this mode.

        ASYMMETRICAL: Asymmetrical quantization, offset is activated in this mode.

        POWER_OF_2: Power-of-2 quantization, scale must be pow(2, k) in this mode.

        DYNAMIC: Dynamic Activation Quantization, scale is computed on the fly.

    ATTENTION: Not all combinations of all 8 QuantizationProperty are valid, see QuantizationPolicy.__check_valid
    ATTENTION: QuantizationPolicy is read-only, user can only assign its value when created, the only interface of
        QuantizationPolicy is function QuantizationPolicy.has_property.
    """

    def __init__(self, policy: int) -> None:
        if not QuantizationPolicy.__check_valid(policy):
            raise ValueError(
                'invalid quantization pattern, valid partterns are listed in '
                'esp_ppq.core.OperationQuantizationPolicy.__check_valid'
            )
        self._policy = policy

    def has_property(self, property: QuantizationProperty) -> bool:
        return (self._policy & property.value) != 0

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, QuantizationPolicy):
            raise TypeError('Can only compare QuantizationPolicy object with another QuantizationPolicy object.')
        return self._policy == o._policy

    @classmethod
    def __check_valid(cls, policy):
        return policy in {
            # Standard Int Quantization
            QuantizationProperty.ASYMMETRICAL | QuantizationProperty.LINEAR | QuantizationProperty.PER_CHANNEL,
            QuantizationProperty.ASYMMETRICAL | QuantizationProperty.LINEAR | QuantizationProperty.PER_TENSOR,
            QuantizationProperty.SYMMETRICAL | QuantizationProperty.LINEAR | QuantizationProperty.PER_CHANNEL,
            QuantizationProperty.SYMMETRICAL | QuantizationProperty.LINEAR | QuantizationProperty.PER_TENSOR,
            QuantizationProperty.ASYMMETRICAL
            | QuantizationProperty.LINEAR
            | QuantizationProperty.PER_TENSOR
            | QuantizationProperty.POWER_OF_2,
            QuantizationProperty.SYMMETRICAL
            | QuantizationProperty.LINEAR
            | QuantizationProperty.PER_TENSOR
            | QuantizationProperty.POWER_OF_2,
            QuantizationProperty.ASYMMETRICAL
            | QuantizationProperty.LINEAR
            | QuantizationProperty.PER_CHANNEL
            | QuantizationProperty.POWER_OF_2,
            QuantizationProperty.SYMMETRICAL
            | QuantizationProperty.LINEAR
            | QuantizationProperty.PER_CHANNEL
            | QuantizationProperty.POWER_OF_2,
            # Low Precision Float Quantization
            # QuantizationProperty.SYMMETRICAL | QuantizationProperty.FLOATING | QuantizationProperty.PER_CHANNEL,
            # QuantizationProperty.SYMMETRICAL | QuantizationProperty.FLOATING | QuantizationProperty.PER_TENSOR,
            # QuantizationProperty.ASYMMETRICAL | QuantizationProperty.FLOATING | QuantizationProperty.PER_CHANNEL,
            # QuantizationProperty.ASYMMETRICAL | QuantizationProperty.FLOATING | QuantizationProperty.PER_TENSOR,
            QuantizationProperty.SYMMETRICAL
            | QuantizationProperty.FLOATING
            | QuantizationProperty.PER_CHANNEL
            | QuantizationProperty.POWER_OF_2,
            QuantizationProperty.SYMMETRICAL
            | QuantizationProperty.FLOATING
            | QuantizationProperty.PER_TENSOR
            | QuantizationProperty.POWER_OF_2,
            # QuantizationProperty.ASYMMETRICAL | QuantizationProperty.FLOATING | QuantizationProperty.PER_CHANNEL | QuantizationProperty.POWER_OF_2,
            # QuantizationProperty.ASYMMETRICAL | QuantizationProperty.FLOATING | QuantizationProperty.PER_TENSOR | QuantizationProperty.POWER_OF_2,
            # Dynamic Activation Quantization
            QuantizationProperty.SYMMETRICAL
            | QuantizationProperty.LINEAR
            | QuantizationProperty.PER_TENSOR
            | QuantizationProperty.DYNAMIC,
            QuantizationProperty.ASYMMETRICAL
            | QuantizationProperty.LINEAR
            | QuantizationProperty.PER_TENSOR
            | QuantizationProperty.DYNAMIC,
            QuantizationProperty.SYMMETRICAL
            | QuantizationProperty.LINEAR
            | QuantizationProperty.PER_CHANNEL
            | QuantizationProperty.DYNAMIC,
            QuantizationProperty.ASYMMETRICAL
            | QuantizationProperty.LINEAR
            | QuantizationProperty.PER_CHANNEL
            | QuantizationProperty.DYNAMIC,
            QuantizationProperty.SYMMETRICAL
            | QuantizationProperty.LINEAR
            | QuantizationProperty.PER_TENSOR
            | QuantizationProperty.DYNAMIC
            | QuantizationProperty.POWER_OF_2,
            QuantizationProperty.ASYMMETRICAL
            | QuantizationProperty.LINEAR
            | QuantizationProperty.PER_TENSOR
            | QuantizationProperty.DYNAMIC
            | QuantizationProperty.POWER_OF_2,
            QuantizationProperty.SYMMETRICAL
            | QuantizationProperty.LINEAR
            | QuantizationProperty.PER_CHANNEL
            | QuantizationProperty.DYNAMIC
            | QuantizationProperty.POWER_OF_2,
            QuantizationProperty.ASYMMETRICAL
            | QuantizationProperty.LINEAR
            | QuantizationProperty.PER_CHANNEL
            | QuantizationProperty.DYNAMIC
            | QuantizationProperty.POWER_OF_2,
        }

    def to_dict(self) -> dict:
        """return a dictionary to describe this policy.

        nothing funny.
        """
        return {property.name: self.has_property(property) for property in QuantizationProperty}


class QuantizationStates(Enum):
    """QuantizationStates is a core data structure for PPQ quantization.
    QuantizationStates tells whether a quantization configuration is activated.

    ATTENTION: Changes of QuantizationState will greatly affect execution result.

    For a TensorQuantizationConfig instance, there are 9 available quantization states now.
    Only when state is ACTIVATED or NEGATIVE, corresponding tensor will be quantized during the execution.

    Here we give a brief description of each quantization state:

        INITIAL: given when TensorQuantizationConfig is created, is an initial state of all quantization configuration.

        PASSIVE_INIT: for particular parameter like bias of GEMM(Convolution) and padding value of Pad. Usually it
        does not have an independent quantization scale and offset, while gets quantized with other tensor's configuration.
            For GEMM and Convolution, there bias will be quantized with input scale * weight scale.
            For padding value and clip value, it shares the same scale with its input.
        Those parameters will have a PASSIVE_INIT state when created.

        OVERLAPPED: state OVERLAPPED means there is someone else takes control of current tensor,
        and overlapped tensor quantization configuration will be ignored by optimization algorithms and executor.

        Graph fusion always generate overlapped quantization, for a typical conv - relu fusion,
        the output quantization of convolution will be overlapped by the output tensor of relu.
        State OVERLAPPED cares only about quantization behaviour that cross layers.

        ACTIVATE: means corresponding tensor is ready to be quantized with its configuration.

        PASSIVE: means corresponding tensor is ready to be quantized with its configuration.
            (however its configuration is not stand alone, its scale and offset depends on someone else.)

        BAKED: means corresponding tensor has been pre-quantized, its value can directly
            go forward without quantization.
    """

    INITIAL = 1  # 量化参数刚刚被初始化，当前 config 不生效，数据不能被使用
    ACTIVATED = 4  # 表示当前 config 生效
    BAKED = 2  # 只针对参数量化，表示参数已经被静态量化，当前 config 不生效，数据可以直接使用
    OVERLAPPED = 3  # 表示这一路输入不量化，当前量化信息被父量化信息所覆盖

    PASSIVE_INIT = 6  # 表示这一路输入被动量化，并且刚刚初始化不能被使用
    PASSIVE = 5  # 表示这一路输入被动量化，如 bias, clip value 等，被动量化参数使用其他 TQC 的量化信息完成量化
    PASSIVE_BAKED = 7  # 被动量化且静态量化，当前config不生效，数据可以直接使用
    FP32 = 8  # 表示这一路输入不量化

    SOI = -1  # Legacy State
    DEQUANTIZED = -2  # Legacy State
    DEACTIVED = -3  # Legacy State

    @classmethod
    def is_activated(cls, state) -> bool:
        return state in {QuantizationStates.ACTIVATED, QuantizationStates.PASSIVE}

    @classmethod
    def can_export(cls, state) -> bool:
        return state not in {
            QuantizationStates.INITIAL,
            QuantizationStates.PASSIVE_INIT,
            QuantizationStates.DEQUANTIZED,
            QuantizationStates.DEACTIVED,
        }


class TensorQuantizationConfig(Serializable):
    """
    ## TensorQuantizationConfig(Tensor 量化控制结构体)
    PPQ 使用量化控制结构体描述量化行为，该结构体被定义在 esp_ppq.core.quant 中。截止 PPQ 0.6.6 版本，该结构体由 15 项不同的属性组成。我们将向你介绍这一核心数据结构体的设计构想。

    ### 1. QuantizationPolicy 量化策略
    在 TensorQuantizationConfig 当中，首当其冲地内容是 TQC.policy，这是一个 QuantizationPolicy 对象。
    policy 属性用于描述量化的规则，一个完整的量化策略是由多个量化属性(QuantizationProperty)组合完成的；在 PPQ 中目前我们支持 8 种不同的量化属性，你可以使用以下属性来组合形成自定义的量化规则:

    1. PER_TENSOR: 以 Tensor 为单位完成量化，每个 Tensor 使用一个 scale 和 offset 信息。

    2. PER_CHANNEL: 以 Channel 为单位完成量化，每个 Channel 使用一个 scale 和 offset 信息。

    3. LINEAR: 线性量化，通常的 INT8, INT16 皆属于线性量化，在线性量化的表示中不存在指数位。

    4. FLOATING: 浮点量化，包括 FP8 E4M3, FP8 E5M2, FP16, BF16 等格式，在浮点量化中数据由底数和指数两部分组成。

    5. SYMMETRICAL: 对称量化，在量化计算中不启用 offset。

    6. ASYMMETRICAL: 非对称量化，在量化计算中启用 offset 完成量化偏移。

    7. POWER_OF_2: 限制 scale 取值必须为 2 的整数次幂，这一量化行为多见于端侧以及浮点量化。

    8. DYNAMIC: 启用动态量化策略，对于每一批次的数据，scale 与 offset 都将被动态地计算更新。

    下图解释了浮点量化与线性量化的区别：

    ![image](https://user-images.githubusercontent.com/43309460/199235366-1e83ed97-0731-4e1d-abeb-b7121e3d2a94.png)

    ### 2. 线性量化与相关属性

    线性量化允许与下列属性进行组合：

        QuantizationProperty.ASYMMETRICAL | QuantizationProperty.LINEAR | QuantizationProperty.PER_CHANNEL,
        QuantizationProperty.ASYMMETRICAL | QuantizationProperty.LINEAR | QuantizationProperty.PER_TENSOR,
        QuantizationProperty.SYMMETRICAL | QuantizationProperty.LINEAR | QuantizationProperty.PER_CHANNEL,
        QuantizationProperty.SYMMETRICAL | QuantizationProperty.LINEAR | QuantizationProperty.PER_TENSOR,
        QuantizationProperty.ASYMMETRICAL | QuantizationProperty.LINEAR | QuantizationProperty.PER_TENSOR | QuantizationProperty.POWER_OF_2,
        QuantizationProperty.SYMMETRICAL | QuantizationProperty.LINEAR | QuantizationProperty.PER_TENSOR | QuantizationProperty.POWER_OF_2,
        QuantizationProperty.ASYMMETRICAL | QuantizationProperty.LINEAR | QuantizationProperty.PER_CHANNEL | QuantizationProperty.POWER_OF_2,
        QuantizationProperty.SYMMETRICAL | QuantizationProperty.LINEAR | QuantizationProperty.PER_CHANNEL | QuantizationProperty.POWER_OF_2,

    线性量化是最为常用的数值量化方法，有些时候我们也称其为均匀量化，在线性量化中，量化操作的计算方法为：

    - Unscaled FP32 = (FP32 / scale) - offset
    - INT8 = Clip(Round(Unscale FP32), quant_min, quant_max)
    - Dequantized FP32 = (INT8 + offset) * scale

    其中 Round 函数行为由 TQC.rounding(RoundingPolicy) 属性确定，PPQ 支持 7 种不同的取整策略，其中 ROUND_HALF_EVEN 是最常见的取整策略，关于取整策略的详细讨论可以参考 https://en.wikipedia.org/wiki/Rounding

    quant_min, quant_max 分别由 TQC.quant_min, TQC.quant_max 属性确定，对于线性量化而言他们是整数，通常为[-128, 127]。部分框架使用 [-127, 127] 作为截断值，在部分场景下如此定义将有优势，但在 Onnx 的 Q/DQ 算子定义中不允许使用 [-127, 127] 作为截断。

    PPQ 可以模拟 1-64 bit 的任意位宽量化，但若以部署为目的，不建议使用 8 bit 之外的配置。用户须知高位宽量化可能造成 Scale 过小，以至于浮点下溢出。

    ### 3. 浮点量化与相关属性

    浮点量化允许与下列属性进行组合：

        QuantizationProperty.SYMMETRICAL | QuantizationProperty.FLOATING | QuantizationProperty.PER_CHANNEL | QuantizationProperty.POWER_OF_2,
        QuantizationProperty.SYMMETRICAL | QuantizationProperty.FLOATING | QuantizationProperty.PER_TENSOR | QuantizationProperty.POWER_OF_2,

    在浮点量化中，量化函数的计算方法为：

    - Unscaled FP32 = (FP32 / scale)
    - FP8 = Convert(Unscale FP32, quant_min, quant_max)
    - Dequantized FP32 = FP8 * scale

    其中 Convert 函数行为复杂，其转换过程分为三种不同的情况：
    - 当 Unscaled FP32 大于 quant_max，或者小于 quant_min，则直接进行截断
    - 当 Unscaled FP32 幅值大于 FP8 能够表达的最小值，此时需要移去多余的底数位，并对底数进行四舍五入
    - 当 Unscaled FP32 数据小于规范化 FP8 能够表达的最小值，此时浮点下溢出，此时我们计算 FP8 = Round(Unscaled FP32 / FP8_min) * FP8_min

    其中 FP8_min 是非规格化 FP8 能够表达的最小值。对于 FP8 E4M3 标准而言，其能表示的最大值为 448.0，最小值为 -448.0。

    quant_min, quant_max 分别由 TQC.quant_min, TQC.quant_max 属性确定，对于 FLOATING 量化，我们引入一个新的属性 TQC.exponent_bits(int)。使用这个属性来指定总位宽中有多少数位用于表示指数(相应地，底数位为总位宽-指数位-1)。

    在浮点量化中，尺度因子的选取对量化效果的影响不大，因此用户可以使用 constant 校准策略(见 esp_ppq.quantization.observer)将所有尺度因子设置为1。

    关于浮点量化的具体细节可以参考 [本文](https://zhuanlan.zhihu.com/p/574825662)

    ### 4. 其他量化控制属性

    1. TQC.num_of_bits(int)：量化位宽，对于 INT8, FP8 量化，量化位宽为 8。对于 INT16, FP16 量化，量化位宽为16。

    2. TQC.state(QuantizationStates): 量化状态，在 PPQ 中目前有共计 8 种不同的量化状态，该属性极大地丰富了 PPQ 量化信息的语义，使得我们能够更加灵活地控制量化行为。该属性可以被用于切换 量化 / 非量化 状态；执行量化联合定点；执行参数烘焙。

    3. TQC.channel_axis(int): 量化轴，对于 PER_CHANNEL 量化，使用这个属性来指定沿着那一维度展开量化，如执行 Per-tensor 量化，该属性被忽略，用户可以将其设置为 None。

    4. TQC.observer_algorithm(str): observer 算法，其中 observer 是用于确定 scale 和 offset 的对象，使用这个属性指明要使用何种类型的 observer 确定 scale 和 offset

    5. TQC.dominator(TensorQuantizationConfig): 一个指向父量化信息的指针。在 PPQ 中 TQC 与 TQC 之间并不是独立的，他们之间可以存在父子关系。所有子量化信息与父量化信息共享 scale 和 offset

    6. TQC.visiblity(QuantizationVisibility): 导出可见性，使用这个属性来告知 ppq 的导出器是否需要导出当前的 TQC。

    ### 5. 量化控制结构体的初始化

    TensorQuantizationConfig 是 PPQ 中的核心数据结构，它总是由 Quantizer 对象完成创建的：

        # 下面这段代码为一个指定的算子创建了相应的 Tensor Quantization Config
        quantizer = PFL.Quantizer(platform=TargetPlatform.TRT_FP8, graph=graph) # 取得 TRT_FP8 所对应的量化器
        quantizer.quantize_operation(op_name = op.name, platform = dispatching[op.name])

    在 PPQ 当中，Quantizer 的职责即是为算子初始化他们的量化控制结构体。不同的量化器将按照不同的规则创建控制结构体，如 TRT_FP8 所对应的量化器 只会为了 Conv, Gemm 算子创建量化信息，要求他们的输入按照对称-浮点-Per Channel的方式完成量化。而 DSP_INT8 所对应的量化器为几乎所有算子创建量化信息，要求他们按照非对称-线性-Per Tensor的方式完成量化。

    用户可以手动创建量化控制结构体，使用 esp_ppq.lib 中的接口：

        # 创建一个默认的线性量化控制结构体(对称, per-tensor)
        from esp_ppq.lib import LinearQuantizationConfig
        TQC = LinearQuantizationConfig()

        # 创建一个默认的浮点量化控制结构体(FP8 E4M3)
        from esp_ppq.lib import FloatingQuantizationConfig
        TQC = FloatingQuantizationConfig()

    ### 6. 量化控制结构体的校准

    绝大部分的 TensorQuantizationConfig 在完成初始化之后都无法使用-他们的 scale 与 offset 均为空值，且 Quantizer 在初始化他们时会将其状态(TQC.state)置为 INITIAL，处于这个状态的量化信息在计算过程中不会被启用。

    我们必须送入一定量数据，进行必要 Calibration 操作后才能为网络中的量化信息确定合理的 scale 与 offset 值，这一过程是由种类繁多的 Observer 完成的：

        # PPQ 目前支持 8 种不同的 Observer
        OBSERVER_TABLE = {
            'minmax': TorchMinMaxObserver,
            'kl': TorchHistObserver,
            'percentile': TorchPercentileObserver,
            'mse': TorchMSEObserver,
            'isotone': TorchIsotoneObserver,
            'constant': ConstantObserver,
            'floating': DirectMSEObserver,
            'isotone': ...
        }

    这些 Observer 会负责在网络计算过程中收集必要的统计信息，并为 TQC 的 scale 与 offset 赋予有效的值。在完成一切之后，Observer 还会负责将 TQC 的状态(TQC.state)修改为 ACTIVED。此时量化信息将被正式启用，从而在网络前向传播模拟量化计算。

    关于 Observer 的讨论，可以参考 [本视频](https://www.bilibili.com/video/BV1QF41157aM)

    ### 7. 量化控制结构体的父子链接

    在我们讨论量化时，对于那些存在着多个输入的算子，例如 add, concat，它们的所有输入总是被要求有着相同的 scale。为了表述这种语义，我们为 TQC 添加了 TQC.dominator 属性，这一属性可以指向另一个量化控制结构体。

    假设我们存在两个不同的量化控制结构体 A, B：

    - 语句 A.dominator = B 表示 A 将共享 B 的 scale 与 offset(A.policy, A.num_of_bits等属性仍将使用自己的)。于此同时 A.state 将被修改为 OVERLAPPED(A 将不再启用)
    - 语句 A.master = B 表示 A 将共享 B 的 scale 与 offset(A.policy, A.num_of_bits等属性仍将使用自己的)。于此同时 A.state 将被修改为 PASSIVE(A 将仍然启用，但不具有独立的量化参数)

    如果 A 已经是其他量化结构体 C 的父节点，则上述过程将级联地使得 B 成为 A, C 共同的父节点，A, C 都将共享 B 的 scale 与 offset。

    下图简述了在量化控制结构体的生命周期中，量化状态是如何变迁的（[量化优化过程](https://github.com/openppl-public/ppq/tree/master/ppq/quantization/optim)将负责修改量化控制信息的状态）：

    ![Quantization State](https://user-images.githubusercontent.com/43309460/199236632-ec69ca29-9900-4875-8299-a196546d0dde.png)
    """

    def __init__(
        self,
        policy: QuantizationPolicy,
        rounding: RoundingPolicy = RoundingPolicy.ROUND_HALF_EVEN,
        num_of_bits: int = 8,
        quant_min: int = -127,
        quant_max: int = 128,
        exponent_bits: int = 0,
        scale: Any = None,
        offset: Any = None,
        observer_algorithm: str = None,
        detail: Any = None,
        channel_axis: int = None,
        visibility: QuantizationVisibility = QuantizationVisibility.EXPORT_WHEN_ACTIVE,
        state: QuantizationStates = QuantizationStates.INITIAL,
    ):
        """Create a PPQ Tensor Quantization Configuration Instance.

        Args:
            policy (QuantizationPolicy):
                Quantization policy instance which defines the quantization behaviour from marco view.

            rounding (RoundingPolicy): Rounding policy used in quantization.

            num_of_bits (int): Quantization fraction bits. (2 < num_of_bits < 64)

            exponent_bits (int): Quantization exponent bits. (0 < num_of_bits < 8)
                For Int8 Quantization, num_of_bits = 8 and exponent_bits = 0
                For FP8 Quantization, num_of_bits = 4 and exponent_bits = 4

            quant_min (int): An integer value represents the upper bound(inclusive) of quantized value.

            quant_max (int): An integer value represents the lower bound(inclusive) of quantized value.

            scale (Any):
                Scale of quantized value, for per-tensor quantization policy, we use a single float as its scale,
                while for per-channel quantization policy, it will be an array that contains scales for each channel.

            offset (Any): Quantization offset for ASYMMETRICAL quantization policy,
                it will be set as 0 in SYMMETRICAL quantization schema.

            observer_algorithm (str): A string represents an observing algorithm for this tensor.
                PPQ support 'kl', 'minmax' observer now.

            detail (Any, optional): Only used by PPQ internal logic, detail is used to store some internal data,
                you are not supposed to use it.

            channel_axis (int, optional): Only used in PER_CHANNEL quantization, channel index.

            visiblity (Visiblity): visiblity is the attribute that controls export logic.

            Currently, there are 3 Visiblity level in PPQ:
            if Visiblity == FORCE_EXPORT, ppq exporter will export this TQC
                ignoring state check(even if current TQC has been overrlapped).
            if Visiblity == EXPORT_WHEN_ACTIVD, ppq exporter will export this TQC only when it has been actived.
            if Visiblity == INTERNAL, This TQC will not be exported.

            state (QuantizationStates, optional):
                Defaults to QuantizationStates.INITIAL, see QuantizationStates for more detail.
        """

        assert num_of_bits <= 64, 'Cannot quantize a tensor with more than 64 bits.'
        assert num_of_bits >= 2, 'Cannot quantize a tensor with less than 2 bits.'
        assert exponent_bits <= 8, 'Cannot quantize a tensor with more than 8 bits exponent(fp32 overflow).'
        assert exponent_bits >= 0, 'Cannot quantize a tensor with less than 0 bits exponent.'

        self._policy = policy
        self._exponent_bits = exponent_bits
        self._num_of_bits = num_of_bits
        self._scale = scale
        self._offset = offset
        self.state = state
        self._rounding = rounding
        self._quant_min = quant_min
        self._quant_max = quant_max
        self._channel_axis = channel_axis
        self.observer_algorithm = observer_algorithm
        self.detail = {} if detail is None else detail
        self._dominator = self  # union-find
        self._hash = self.__create_hash()
        self._visibility = visibility
        super().__init__()

    def can_export(self, export_overlapped: bool = EXPORT_OVERLAPPED_CONFIG) -> bool:
        if self.visibility == QuantizationVisibility.INTERNAL:
            return False
        type_check = isinstance(self.scale, torch.Tensor) and isinstance(self.offset, torch.Tensor)
        valid_states = {QuantizationStates.BAKED, QuantizationStates.PASSIVE_BAKED}

        if export_overlapped:
            valid_states.add(QuantizationStates.OVERLAPPED)
        state_check = QuantizationStates.is_activated(self.state) or self.state in valid_states

        if state_check or self.visibility == QuantizationVisibility.FORCE_EXPORT:
            if type_check:
                return True
        return False

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, TensorQuantizationConfig):
            raise TypeError(
                'Can only compare TensorQuantizationConfig object with another TensorQuantizationConfig object.'
            )
        return self._hash == o._hash

    def __str__(self) -> str:
        return f'PPQ TensorQuantizationConfig({self.__hash__()})'

    _hash_seed = int(time.time())

    @staticmethod
    def __create_hash():
        TensorQuantizationConfig._hash_seed = (0x343FD * TensorQuantizationConfig._hash_seed + 0x269EC3) % (2 << 31)
        return TensorQuantizationConfig._hash_seed

    def __hash__(self) -> int:
        return self._hash

    def is_same_scheme(self, o: object) -> bool:
        if not isinstance(o, TensorQuantizationConfig):
            raise TypeError(
                'Can only compare TensorQuantizationConfig object with another TensorQuantizationConfig object.'
            )
        return (
            self.quant_max == o.quant_max
            and self.quant_min == o.quant_min
            and self.policy == o.policy
            and self.num_of_bits == o.num_of_bits
            and self.exponent_bits == o.exponent_bits
            and self.channel_axis == o.channel_axis
            and self.rounding == o.rounding
        )

    @property
    def dominated_by(self):
        """dominated_by is a crucial feature for tensor quantization
        configuration in esp_ppq. This property is actually maintained by union-
        find set data structure.

        Every tensor quantization configuration(A) is created with dominated_by = self, and only when
            it is overlapped by other configuration(B), it shall set A.dominated_by = B.
            Setting A.dominated_by = B also makes A, B as a quantization group.
            (quantization state of A is always set as OVERLAPPED here)

        So to say every tensor quantization configuration with dominated_by != self is overrlaped by
            other quantization configuration. When a tensor quantization configuration is overlapped,
            it means this tensor is already been quantized with another quantization configuration,
            and there is no need to be quantized with this configuration anymore.

        PPQ use this property to find root configuration for each configuration group,

        Returns:
            [TensorQuantizationConfig]: root configuration of this quantization group.

        ATTENTION: This configuration is invalid when self.dominated_by != self.
        """
        if self._dominator == self:
            return self
        else:
            root = self._dominator.dominated_by
            self._dominator = root
            return root

    @dominated_by.setter
    def dominated_by(self, o):
        assert isinstance(o, TensorQuantizationConfig), 'Can only set this attribute with another tensor config.'
        if o._hash == self._hash:
            raise ValueError('Error with TQC.dominated_by = o: o must not equal to TQC its self.')
        root, dominator = self.dominated_by, o.dominated_by
        if self == dominator:
            raise ValueError(
                'Can not Assign Dominator like this, '
                'Circular reference was detected. Son TQC can not dominate its Father.'
            )
        assert isinstance(root, TensorQuantizationConfig)
        if dominator != root:
            root._dominator = dominator
            self._dominator = dominator
            root.state = QuantizationStates.OVERLAPPED
            self.state = QuantizationStates.OVERLAPPED

    @property
    def master_by(self):
        if self._dominator == self:
            return self
        else:
            root = self._dominator.dominated_by
            self._dominator = root
            return root

    @master_by.setter
    def master_by(self, master):
        if not isinstance(master, TensorQuantizationConfig):
            raise TypeError(
                f'Error with TQC.master_by(o): o must be another Tensor Quantization Config, '
                f'however {type(master)} was given.'
            )
        if master._hash == self._hash:
            raise ValueError('Error with TQC.dominated_by = o: o must not equal to TQC its self.')
        self._dominator = master
        if master.scale is not None and master.offset is not None:
            self.state = QuantizationStates.PASSIVE
        else:
            self.state = QuantizationStates.PASSIVE_INIT

    def is_revisable(self):
        return self.dominated_by == self and self.state in {
            QuantizationStates.ACTIVATED,
            QuantizationStates.FP32,
            QuantizationStates.FP32,
            QuantizationStates.INITIAL,
            QuantizationStates.FP32,
            QuantizationStates.PASSIVE,
            QuantizationStates.PASSIVE_INIT,
        }

    @property
    def visibility(self) -> QuantizationVisibility:
        """Export Visibility of this TQC.

        * QuantizationVisibility.EXPORT_WHEN_ACTIVE - Export this TQC when it is active.

        * QuantizationVisibility.FORCE_EXPORT - Force Export this TQC.

        * QuantizationVisibility.INTERNAL - Never Export this TQC.

        """
        return self._visibility

    @visibility.setter
    def visibility(self, visiblity: QuantizationVisibility):
        self._visibility = visiblity

    @property
    def scale(self) -> torch.Tensor:
        """Get Quantization Scale of this TQC.

        If current TQC is dominated by other, return father TQC's scale instead.
        """
        if self.dominated_by == self:
            return self._scale
        else:
            return self.dominated_by.scale

    @property
    def offset(self) -> torch.Tensor:
        """Get Quantization Offset of this TQC.

        If current TQC is dominated by other, return father TQC's offset instead.
        """
        if self.dominated_by == self:
            return self._offset
        else:
            return self.dominated_by.offset

    @property
    def policy(self) -> QuantizationPolicy:
        """Get Quantization Policy of this TQC."""
        return self._policy

    @property
    def num_of_bits(self) -> int:
        """Get bit-width of this TQC."""
        return self._num_of_bits

    @property
    def rounding(self) -> RoundingPolicy:
        """Get Rounding Policy of this TQC."""
        return self._rounding

    @property
    def quant_min(self) -> int:
        """Get minimum quant value of this TQC."""
        return self._quant_min

    @property
    def quant_max(self) -> int:
        """Get maximum quant value of this TQC."""
        return self._quant_max

    @property
    def exponent_bits(self) -> int:
        """Get exponent bit-width of current TQC.

        num_of_bits = exponent_bits + mantissa_bits
        """
        return self._exponent_bits

    @property
    def mantissa_bits(self) -> int:
        """Get mantissa bit-width of current TQC.

        num_of_bits = exponent_bits + mantissa_bits
        """
        # there is one bit for sign.
        return self.num_of_bits - self._exponent_bits - 1

    @property
    def channel_axis(self) -> int:
        """Get Quantization Axis, For Per-tensor Quantization, it returns None."""
        return self._channel_axis

    @scale.setter
    def scale(self, value: Any):
        if not self.is_revisable():
            raise PermissionError(
                'Can not change scale of this tensor quantization configuration now. '
                'It has been overlapped or has an inactive state. '
                'Due to it is not a active config, any change of this configuration is not allowed.'
            )
        else:
            self._scale = value

    @offset.setter
    def offset(self, value: Any):
        if not self.is_revisable():
            raise PermissionError(
                'Can not change offset of this tensor quantization configuration now. '
                'It has been overlapped or has an inactive state. '
                'Due to it is not a active config, any change of this configuration is not allowed.'
            )
        else:
            self._offset = value

    @policy.setter
    def policy(self, policy: QuantizationPolicy):
        self._policy = policy

    @num_of_bits.setter
    def num_of_bits(self, bits: int):
        self._num_of_bits = bits

    @rounding.setter
    def rounding(self, policy: RoundingPolicy):
        self._rounding = policy

    @quant_min.setter
    def quant_min(self, min: int):
        self._quant_min = min

    @quant_max.setter
    def quant_max(self, max: int):
        self._quant_max = max

    @exponent_bits.setter
    def exponent_bits(self, bits: int):
        if not self.policy.has_property(QuantizationProperty.FLOATING):
            raise PermissionError(
                'Can not change property: exponent bits for this TQC. '
                'self.policy.has_property(QuantizationProperty.FLOATING) == False.'
            )
        self._exponent_bits = bits

    @channel_axis.setter
    def channel_axis(self, channel_axis: int):
        if not self.policy.has_property(QuantizationProperty.PER_CHANNEL):
            raise PermissionError(
                'Can not change property: quantization channel axis for this TQC. '
                'self.policy.has_property(QuantizationProperty.PER_CHANNEL) == False.'
            )
        self._channel_axis = channel_axis

    def copy(self):
        """Create a tensor config from this one, keep policy and state
        unchanged.

        if there is an non-empty scale and offset, they will be cloned too.
        """
        scale, offset = None, None
        if self.scale is not None:
            if isinstance(self.scale, torch.Tensor):
                scale = self.scale.clone()
            else:
                scale = self.scale
        if self.offset is not None:
            if isinstance(self.offset, torch.Tensor):
                offset = self.offset.clone()
            else:
                offset = self.offset
        config = TensorQuantizationConfig(
            policy=self.policy,
            rounding=self.rounding,
            num_of_bits=self.num_of_bits,
            quant_min=self.quant_min,
            quant_max=self.quant_max,
            scale=scale,
            offset=offset,
            observer_algorithm=self.observer_algorithm,
            detail=self.detail.copy(),
            state=self.state,
            exponent_bits=self.exponent_bits,
            channel_axis=self.channel_axis,
            visibility=self.visibility,
        )
        if self.state == QuantizationStates.OVERLAPPED:
            config._dominator = self._dominator
        return config


class ChannelwiseTensorQuantizationConfig(TensorQuantizationConfig):
    """Legacy Class Since PPQ 0.6.6, Use TensorQuantizationConfig Instead."""

    def __init__(
        self,
        policy: QuantizationPolicy,
        rounding: RoundingPolicy,
        num_of_bits: int,
        exponent_bits: int,
        quant_min: int,
        quant_max: int,
        scale: Any,
        offset: Any,
        observer_algorithm: str,
        state: QuantizationStates,
        channel_axis: int,
        detail: dict = {},
    ):
        ppq_warning(
            'ChannelwiseTensorQuantizationConfig is now obsolescent(Since PPQ 0.6.6), '
            'use TensorQuantizationConfig Instead.'
        )
        if policy.has_property(QuantizationProperty.PER_TENSOR):
            raise TypeError(
                'Can not assign QuantizationProperty.PER_TENSOR policy '
                'to a Channel-wise Tensor Quantization Config instance.'
            )
        super().__init__(
            policy=policy,
            num_of_bits=num_of_bits,
            quant_min=quant_min,
            quant_max=quant_max,
            scale=scale,
            offset=offset,
            observer_algorithm=observer_algorithm,
            detail=detail,
            state=state,
            rounding=rounding,
            exponent_bits=exponent_bits,
        )
        self.channel_axis = channel_axis

    @classmethod
    def convert_from_tensor_config(
        cls,
        convert_from: TensorQuantizationConfig,
        scale: Iterable = None,
        offset: Iterable = None,
        channel_axis: int = 1,
    ):
        if scale is None:
            scale = convert_from.scale
        if offset is None:
            offset = convert_from.offset
        this = ChannelwiseTensorQuantizationConfig(
            policy=convert_from.policy,
            num_of_bits=convert_from.num_of_bits,
            quant_min=convert_from.quant_min,
            quant_max=convert_from.quant_max,
            scale=scale,
            offset=offset,
            observer_algorithm=convert_from.observer_algorithm,
            detail=convert_from.detail.copy(),
            state=convert_from.state,
            channel_axis=channel_axis,
            rounding=convert_from.rounding,
            exponent_bits=convert_from.exponent_bits,
        )
        return this

    def copy(self):
        config = super().copy()
        return self.convert_from_tensor_config(
            config, scale=config.scale, offset=config.offset, channel_axis=self.channel_axis
        )


class OperationQuantizationConfig(Iterable):
    """OperationQuantizationConfig serves as a collection of tensor
    quantization configuration.

    See TensorQuantizationConfig for more information.
    """

    def __init__(
        self,
        input_quantization_configs: List[TensorQuantizationConfig] = None,
        output_quantization_configs: List[TensorQuantizationConfig] = None,
        is_positive_quant_op: bool = True,
    ):
        """Create an operation quantization configuration.

        Args:
            input_quantization_configs (List[TensorQuantizationConfig], optional):
                a list contains all configuration of all input variables.

            output_quantization_configs (List[TensorQuantizationConfig], optional):
                a list contains all configuration of all output variables.

            ATTENTION: whether a variable is gonna to be quantized or not, it must have a quantization configuration.

            is_positive_quant_op (bool, optional): [description]. Defaults to True.
                some operations are passively quantized, such as Maxpooling, Padding.
                For those operations, set this property as False, PPQ will use this property to optimize your graph.
        """
        self.input_quantization_config = self.__check_famliy_config(input_quantization_configs)
        self.output_quantization_config = self.__check_famliy_config(output_quantization_configs)
        self.is_active_quant_op = is_positive_quant_op

    def export(self) -> str:
        raise Exception('Implement this first')

    def __check_famliy_config(self, famliy_configs):
        for famliy_config in famliy_configs:
            if not isinstance(famliy_config, TensorQuantizationConfig):
                raise TypeError(
                    f'You are trying to set famliy quantization config of {str(self)}, '
                    f'However your input is invalid, except one TensorQuantizationConfig object, '
                    f'while a {type(famliy_config)} was given.'
                )
        return famliy_configs

    def __str__(self) -> str:
        return f'Inputs config: {self.input_quantization_config}, Outputs config {self.output_quantization_config}'

    def __iter__(self) -> TensorQuantizationConfig:
        return (self.input_quantization_config + self.output_quantization_config).__iter__()

    def copy(self):
        """Create an operation config from this one, keep policy and state
        unchanged.

        if this one has an non-empty scale or offset, they will be cloned too.
        """
        return OperationQuantizationConfig(
            input_quantization_configs=[_.copy() for _ in self.input_quantization_config],
            output_quantization_configs=[_.copy() for _ in self.output_quantization_config],
            is_positive_quant_op=self.is_active_quant_op,
        )



"""

                                                              ...::iirii:..
                P7Ysi.  iS7u.                ..:::irri:rPBBBBBBBBBBBBBBBBBBBBBBBg57.
                B   :Js:B7 .PS        :sEBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBDv
                B      .:     u  :5BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB2
                DQ            SBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBQBBBBBBBBBBBBBBBBBBBBBBBBBBRi
           JZ  ..i:.      .dBBBBBBBBBBBBBBBBBBBBBBBg7          .BQBBBBBBBBBBBBBBBBBBBBBBBBBBBBBD7
            YQ.         vBBBBBBBBBBBBBBBBBBBBBBMY.              QBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB7
              2D      YBBBBBBBBBBBBBBBBBBBBZ7                   BBBBBBBBBBBBBBr SBBBBBBBBBBBBBBBBBBBBM:
               :Bi  UBBBBBBBBBBBBBBBBBPv.                       BBBBBBBBBBBBj    .BBBBBBBBBBBBBBBBBBBBBBY
                 ZQBBBBBBBBBBBBBBE                             DBBBBBBBBBBS        .PBBBBBBBBBBBBBBBBBBBBBu
                ZBBBBBBBBBBBBBBBZ                             EBBBBBBBBb:             :PBBBBBBBBBBBBBBBBBBBB
             .iBBBBBBBBBBBBBBBBB          ..:i:.            .BBBBBQKr                     7gBBBBBBBBBBBBBBBBB
             gBBBBBBBBBBBBBBBBQ      .irri::...           IR2r.                   rir7j17.    :LBBBBQRQBBBBBBB
             BBBBBBBBBBBBBBBB:     rur.                 .i.                              iKg7     vBBBBBBBBBBB:
            BBBBBBBBBBBBBBP.      Xr    ..:r7rrr.                              .::LKIvrr.   5Br     :BBBBBBBBB1
           bBBBBBBBBBBZ7.            :Yr.     :dBBP                         .JYi.      :PBE   E       BBBBBBBBr
          jBBBBBBBBBB              :Q7    .rv:   vBb                        Qi            BB:         :BBBBBBB.
          BBBBBBBBBBr              B.   BBBBBBBBQ                                5BBBBBQi  dB          IBBBBQB5
         :BBBBBBBBBB                   JBBBBBBBBB:                              BBBBBBBBB:               rBBBBB
          BBBBBBBBB1                    XBBBBBBU                                ZBBBBBBBB.                EBBBB7MQ
          BBBBBBBBB7                                                              .LU1r.                  iBBBb  JBv
          BBBBBBBBg                  5D7                         ..                        .              :BBB7   rB
          qB: gBBBK           i  : .  .Y21JJL7i.                 ..B:             ULi::iirr:..            .BBu  :  B7
        gd Ui  QBBE            .   :.                              r.              .....    :i. :.        PBR     .B.
       7B .    iBBB                                                                             .        .B 7r    BL
       BB .     UBBv                                                                                     BP .   rBr
       MB      q  .                                                                                     XS    rDj:
       .BB    .::. :                  iLv7ri:.                                      ...                ..  :YUr
         BB7     .:U                   ...irvUXPP27:                      ..:71SqP2Is7i              Ui::rr:
          .BBg:                                  .rLr:r7i..        .:irrrii:..                      Bg
             7QBE7.                                     .:r77v .rri:.                             rB7
                :JdDE277S.                                   ..r                                rDK
                      . rBBi                                                                .is5r
                          :gBEr                                                          :7jL:
                             :SQMSr.                                                 .:7r:.
                                 :YKPbur::..                                 ...irrqBq.
                                      :KgBQU1UUJr:          i7777vsv777v777ri:...    rdQJ.
                                    7qj:                          ..              .     7QBr
                                  rP:                                                      bB

Sensetime PPL Quant Tool(PPQ)

Do not modify codes within this directory if not necessary.
"""

from typing import Any, Callable, Dict, Iterable, List, Set, Text, Union

from .common import *
from .config import PPQ_CONFIG
from .data import (
    DataType,
    OperationMeta,
    TensorMeta,
    convert_any_to_numpy,
    convert_any_to_python_primary_type,
    convert_any_to_string,
    convert_any_to_torch_tensor,
    convert_primary_type_to_list,
)
from .defs import (
    SingletonMeta,
    empty_ppq_cache,
    ppq_debug_function,
    ppq_file_io,
    ppq_info,
    ppq_legacy,
    ppq_quant_param_computing_function,
    ppq_warning,
)
from .ffi import CUDA
from .quant import (
    ChannelwiseTensorQuantizationConfig,
    NetworkFramework,
    OperationQuantizationConfig,
    QuantizationPolicy,
    QuantizationProperty,
    QuantizationStates,
    QuantizationVisibility,
    RoundingPolicy,
    TargetPlatform,
    TensorQuantizationConfig,
)
from .storage import Serializable, ValueState, is_file_exist, open_txt_file_from_writing

print(
    r"""
    ___________ ____        ____  ____  ____
   / ____/ ___// __ \      / __ \/ __ \/ __ \
  / __/  \__ \/ /_/ /_____/ /_/ / /_/ / / / /
 / /___ ___/ / ____/_____/ ____/ ____/ /_/ /
/_____//____/_/         /_/   /_/    \___\_\

"""
)



from .default import DEFAULT_BACKEND_TABLE

PPL_DSP_BACKEND_TABLE = DEFAULT_BACKEND_TABLE.copy()


# When you trying to implement a custimized function for ppl_dsp platform
# Be aware that you can just overwrite part of DEFAULT_DISPATCHING_TABLE
# rather than rewrite all dispatching table.
# here an example was given: Sample_Forward
def Sample_Forward():
    return None


PPL_DSP_BACKEND_TABLE['Sample_Function'] = Sample_Forward



from typing import Union

import torch

from esp_ppq.core import (
    PASSIVE_OPERATIONS,
    OperationQuantizationConfig,
    QuantizationPolicy,
    QuantizationProperty,
    QuantizationStates,
    RoundingPolicy,
    TargetPlatform,
)
from esp_ppq.IR import BaseGraph, GraphCommandProcessor, Operation

from .base import BaseQuantizer


class MNNQuantizer(BaseQuantizer):
    def __init__(self, graph: Union[BaseGraph, GraphCommandProcessor]) -> Union[torch.Tensor, list, dict]:
        super().__init__(graph=graph)
        self._num_of_bits = 8
        self._quant_min = -127
        self._quant_max = +127

    def init_quantize_config(self, operation: Operation) -> OperationQuantizationConfig:
        base_quant_config = self.create_default_quant_config(
            policy=self.quantize_policy,
            rounding=self.rounding_policy,
            op=operation,
            num_of_bits=self._num_of_bits,
            exponent_bits=0,
            quant_max=self._quant_max,
            quant_min=self._quant_min,
            observer_algorithm='percentile',
        )

        if operation.type == 'Conv':
            assert operation.num_of_input > 0, 'Seems you got a Conv layer with no parameters.'

            if operation.inputs[1].is_parameter:
                conv_weight_config = base_quant_config.input_quantization_config[1]
                conv_weight_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_CHANNEL
                )
                conv_weight_config.channel_axis = 0
                conv_weight_config.observer_algorithm = 'minmax'

            if operation.num_of_input > 2:
                bias_config = base_quant_config.input_quantization_config[-1]
                bias_config.state = QuantizationStates.FP32

        if operation.type in PASSIVE_OPERATIONS:
            # Those op are not active op.
            base_quant_config.is_active_quant_op = False
        return base_quant_config

    @property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.MNN_INT8

    @property
    def default_platform(self) -> TargetPlatform:
        return TargetPlatform.FP32

    @property
    def quant_operation_types(self) -> set:
        return {'Conv', 'Add', 'Gemm'}

    @property
    def quantize_policy(self) -> QuantizationPolicy:
        return QuantizationPolicy(
            QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_TENSOR
        )

    @property
    def rounding_policy(self) -> RoundingPolicy:
        return RoundingPolicy.ROUND_HALF_FAR_FORM_ZERO

    @property
    def activation_fusion_types(self) -> set:
        return {'Relu', 'Clip', 'Swish', 'SoftPlus', 'Sigmoid'}



from typing import Dict, Set

from esp_ppq.core import TargetPlatform
from esp_ppq.IR import BaseGraph

from .base import GraphDispatcher


class AllinDispatcher(GraphDispatcher):
    """Graph Dispatcher cuts a graph into parts, each part of graph will
    dispatch to a specific platform for further execution and quantization.
    ATTENTION: this dispatcher will enable all ops in quant_types to quant_platform.
    """

    def __init__(self, graph: BaseGraph) -> None:
        super().__init__()
        self.graph = graph

    def dispatch(
        self,
        quant_types: Set[str],
        quant_platform: TargetPlatform = TargetPlatform.UNSPECIFIED,
        fp32_platform: TargetPlatform = TargetPlatform.FP32,
        SOI_platform: TargetPlatform = TargetPlatform.SOI,
        **kwargs,
    ) -> Dict[str, TargetPlatform]:
        """
            We assume all ops in origin model can be quant.
            This is suitable for some npu platform.
        Args:
            graph (BaseGraph): graph object which going to be dispatched by this dispatcher.
            quant_types(Set[str]): all quantable types for given platforms.
            quant_platform (TargetPlatform):
                platform object where quantable parts will goes to.
            fp32_platform (TargetPlatform):
                platform object where SOI parts will goes to.
            SOI_platform (TargetPlatform):
                platform object where remaining parts will goes to.
        Returns:
            Dict[str, TargetPlatform]: [description]
        """
        graph = self.graph

        dispatching_table = {}
        for op in graph.operations.values():
            if op.type in quant_types:
                dispatching_table[op.name] = TargetPlatform.UNSPECIFIED
            else:
                dispatching_table[op.name] = TargetPlatform.FP32

        return dispatching_table



from typing import Dict, List

import onnx
import torch
from onnx import helper

from esp_ppq.core import (
    GRAPH_OPSET_ATTRIB,
    PPQ_CONFIG,
    QuantizationProperty,
    QuantizationStates,
    QuantizationVisibility,
    TensorQuantizationConfig,
    convert_any_to_torch_tensor,
    ppq_warning,
)
from esp_ppq.IR import BaseGraph, Operation, QuantableOperation, QuantableVariable, Variable
from esp_ppq.quantization.qfunction.linear import PPQLinearQuant_toInt
from esp_ppq.utils.round import ppq_tensor_round

from .onnx_exporter import OP_CONVERTERS, OnnxExporter, OperationExporter


class QDQHelper:
    """Helper class for processing onnx qdq format"""

    @staticmethod
    def TQC_Exportable_Check(TQC: TensorQuantizationConfig, bounded_var: Variable) -> bool:
        if not TQC.can_export():
            return False

        if TQC.visibility == QuantizationVisibility.INTERNAL:
            return False
        if TQC.num_of_bits == 8 and TQC.policy.has_property(QuantizationProperty.LINEAR):
            if TQC.policy.has_property(QuantizationProperty.ASYMMETRICAL):
                range_check = TQC.quant_max <= 255 and TQC.quant_min >= 0
            else:
                range_check = TQC.quant_max <= 127 and TQC.quant_min >= -128
        else:
            range_check = True

        if not range_check:
            ppq_warning(
                f'Is it not safe to export TQC({bounded_var.name}) to Onnx, '
                f'INT8 value range must be [-128, 127] or [0, 255], '
                f'however [{TQC.quant_min, TQC.quant_max}] was given.'
            )
            return False
        return True


class ONNXRUNTIMExporter(OnnxExporter):
    """
    PPQ 可以将 TQC 中的量化信息导出为 Onnx QDQ 节点。

    对于 INT8 量化，PPQ 可以导出符合 Onnx 量化要求的量化信息，这要求

        * TQC.quant_min   = -128
        * TQC.quant_max   = 127
        * TQC.num_of_bits = 8

    不符合规范的量化信息将不会被导出，并且 PPQ 将给出警告。

    Onnx QDQ 格式具有较高的灵活性，因此 PPQ 与推理框架可能在一些细节格式问题上出现分歧。
    例如对于节点 A 而言，假设它有两个下游节点 B, C。此时 QDQ 节点可以插入在 A 的输出端，也可以分别插入在 B, C 的输入端。
    因此导出模型的格式可能存在不一致，这也有可能导致推理框架解析模型时发生错误。出现上述情况时，用户需要手动对导出逻辑进行干预。

    对于激活值非对称量化的策略而言，激活函数 relu, clip 的功能可以被 QDQ 节点代替，因此通常情况下它们是可省的。
    PPQ 会自动探测上述情况，在确保省略激活函数不会对网络结果产生影响时移除非对称量化中的激活函数。

    PPQ 会移除冗余的 QDQ 节点，即前后相邻的两组 Q-DQ 节点，如果它们的量化参数一致，则只会导出上游节点的量化信息。

    对于浮点量化而言，PPQ 会导出自定义节点 QuantizeFloating, DequantizeFloating， 这些节点不被推理框架识别，仅作可视化用途。

    任何导出器的导出逻辑都是原地进行的，它们将对传入的计算图对象进行原地修改，因此在导出之前你需要手动克隆计算图。
    """

    def __init__(self, removed_activation_types: List[str] = ['Relu', 'Clip']) -> None:
        super().__init__()
        self.removed_activation_types = removed_activation_types

    def infer_qtype(self, config: TensorQuantizationConfig):
        offset_dtype, value_dtype = torch.int8, torch.int8
        if config.policy.has_property(QuantizationProperty.ASYMMETRICAL):
            offset_dtype = torch.uint8
            value_dtype = torch.uint8
        if config.num_of_bits > 8:
            offset_dtype = torch.int32
            value_dtype = torch.int32
        return offset_dtype, value_dtype

    def insert_quantize_node(
        self, graph: BaseGraph, var: Variable, config: TensorQuantizationConfig, op: Operation
    ) -> Operation:
        """
        Insert a Quantize Node on given variable, according to given TensorQuantizationConfig.
        There is two basic type of Quantize Node: QuantizeLinear and QuantizeFloating.
        """
        if config.policy.has_property(QuantizationProperty.LINEAR):
            # Following code will export Linear Quantization Config
            # That is for FP32 -> INT
            offset_dtype, value_type = self.infer_qtype(config)
            scale = convert_any_to_torch_tensor(config.scale.clone(), dtype=torch.float32)
            offset = ppq_tensor_round(config.offset.clone()).type(offset_dtype)

            created = graph.create_operation(op_type='QuantizeLinear', attributes={})
            if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
                created.attributes['axis'] = config.channel_axis
            else:
                created.attributes['axis'] = None

            if var in op.inputs:
                graph.insert_op_before(A=created, B=op, input_idx=op.inputs.index(var))
            elif var in op.outputs:
                graph.insert_op_after(A=created, B=op, output_idx=op.outputs.index(var))
            else:
                raise ValueError(f'Unexpected Error in Exporting Op {op.name}({op.type}).')

            graph.create_variable(name=None, value=scale, is_parameter=True, dest_ops=[created])
            graph.create_variable(name=None, value=offset, is_parameter=True, dest_ops=[created])

            created.outputs[0].dtype = value_type
            created.outputs[0].shape = var.shape
            created.inputs[0].shape = var.shape
            return created

        elif config.policy.has_property(QuantizationProperty.FLOATING):
            # Following code will export Linear Quantization Config
            # That is for FP32 -> FP8
            scale = convert_any_to_torch_tensor(config.scale.clone(), dtype=torch.float32)
            offset = convert_any_to_torch_tensor(config.offset.clone(), dtype=torch.float32)

            created = graph.create_operation(
                op_type='QuantizeFloating',
                attributes={
                    'min': config.quant_min,
                    'max': config.quant_max,
                    'exponent': config.exponent_bits,
                    'mantissa': config.mantissa_bits,
                },
            )

            if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
                created.attributes['axis'] = config.channel_axis
            else:
                created.attributes['axis'] = None

            if var in op.inputs:
                graph.insert_op_before(A=created, B=op, input_idx=op.inputs.index(var))
            elif var in op.outputs:
                graph.insert_op_after(A=created, B=op, output_idx=op.outputs.index(var))
            else:
                raise ValueError(f'Unexpected Error in Exporting Op {op.name}({op.type}).')

            graph.create_variable(name=None, value=scale, is_parameter=True, dest_ops=[created])
            graph.create_variable(name=None, value=offset, is_parameter=True, dest_ops=[created])

            created.outputs[0].shape = var.shape
            created.inputs[0].shape = var.shape
            return created

        else:
            raise TypeError(
                f'PPQ Can not export quantization information with variable {var.name}, '
                'Unexpected Quantization property.'
            )

    def insert_dequantize_node(
        self, graph: BaseGraph, var: Variable, config: TensorQuantizationConfig, op: Operation
    ) -> Operation:
        """
        Insert a DeQuantize Node on given variable, according to given TensorQuantizationConfig.
        There is two basic type of DeQuantize Node: DeQuantizeLinear and DeQuantizeFloating.
        """
        if config.policy.has_property(QuantizationProperty.LINEAR):
            offset_dtype, value_type = self.infer_qtype(config)
            scale = convert_any_to_torch_tensor(config.scale.clone(), dtype=torch.float32)
            offset = ppq_tensor_round(config.offset.clone()).type(offset_dtype)

            created = graph.create_operation(op_type='DequantizeLinear', attributes={})
            if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
                created.attributes['axis'] = config.channel_axis
            else:
                created.attributes['axis'] = None

            if var in op.inputs:
                graph.insert_op_before(A=created, B=op, input_idx=op.inputs.index(var))
            elif var in op.outputs:
                graph.insert_op_after(A=created, B=op, output_idx=op.outputs.index(var))
            else:
                raise ValueError(f'Unexpected Error in Exporting Op {op.name}({op.type}).')

            graph.create_variable(name=None, value=scale, is_parameter=True, dest_ops=[created])
            graph.create_variable(name=None, value=offset, is_parameter=True, dest_ops=[created])

            created.inputs[0].dtype = value_type
            created.inputs[0].shape = var.shape
            created.outputs[0].shape = var.shape
            created.outputs[0].dtype = torch.float32

            return created

        elif config.policy.has_property(QuantizationProperty.FLOATING):
            scale = convert_any_to_torch_tensor(config.scale.clone(), dtype=torch.float32)
            offset = convert_any_to_torch_tensor(config.offset.clone(), dtype=torch.float32)

            created = graph.create_operation(
                op_type='DequantizeFloating',
                attributes={
                    'min': config.quant_min,
                    'max': config.quant_max,
                    'exponent': config.exponent_bits,
                    'mantissa': config.mantissa_bits,
                },
            )

            if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
                created.attributes['axis'] = config.channel_axis
            else:
                created.attributes['axis'] = None

            if var in op.inputs:
                graph.insert_op_before(A=created, B=op, input_idx=op.inputs.index(var))
            elif var in op.outputs:
                graph.insert_op_after(A=created, B=op, output_idx=op.outputs.index(var))
            else:
                raise ValueError(f'Unexpected Error in Exporting Op {op.name}({op.type}).')

            graph.create_variable(name=None, value=scale, is_parameter=True, dest_ops=[created])
            graph.create_variable(name=None, value=offset, is_parameter=True, dest_ops=[created])

            created.outputs[0].shape = var.shape
            created.inputs[0].shape = var.shape

            return created

        else:
            raise TypeError(
                f'PPQ Can not export quantization information with variable {var.name}, '
                'Unexpected Quantization property.'
            )

    def remove_activation_ops(self, graph: BaseGraph) -> BaseGraph:
        """For Asymmetric Quantization Policy, Activations like Relu & Clip can
        be removed from your network safely. Their function can be replaced by
        quant & dequant operations.

        Those activation is unnecessary for Asymmetric quantized network.

        Args:
            graph (BaseGraph): Processing Graph
            activation_ops (List[Operation]): Removing activations.
        """
        removed_activations = []
        for op in graph.topological_sort():
            if not isinstance(op, QuantableOperation):
                continue
            if op.type in {'Relu', 'Clip'}:
                config = op.config.output_quantization_config[0]
                # Only ASYMMETRICAL quantized activations can be safely removed.
                if config.policy.has_property(QuantizationProperty.SYMMETRICAL):
                    continue

                if not isinstance(config.scale, torch.Tensor):
                    continue
                if not isinstance(config.offset, torch.Tensor):
                    continue

                range_min = (config.scale * (config.quant_min - config.offset)).min().item()
                range_max = (config.scale * (config.quant_max - config.offset)).max().item()

                if op.type == 'Relu':
                    if range_min >= 0:
                        removed_activations.append(op)

                if op.type == 'Clip':
                    if op.num_of_input == 3:
                        clip_min = op.inputs[1].value
                        clip_max = op.inputs[2].value
                        if clip_min is not None:
                            clip_min = clip_min.item()
                        else:
                            clip_min = float('-inf')
                        if clip_max is not None:
                            clip_max = clip_max.item()
                        else:
                            clip_max = float('+inf')

                        if range_min >= clip_min and range_max <= clip_max:
                            removed_activations.append(op)

        # Activation op can only be relu and clip,
        # so it is safe to access op.inputs[0], op.outputs[0] as their input and output.
        for op in removed_activations:
            if not isinstance(op, QuantableOperation):
                continue
            if len(graph.get_upstream_operations(op)) == 0:
                continue
            config = op.config.output_quantization_config[0]

            upstream_op = graph.get_upstream_operations(op)[0]
            if not isinstance(upstream_op, QuantableOperation):
                continue
            if len(graph.get_downstream_operations(upstream_op)) != 1:
                continue
            input_var, input_cfg = op.inputs[0], op.config.input_quantization_config[0]
            if not input_cfg.policy.has_property(QuantizationProperty.ASYMMETRICAL):
                continue

            # PATCH 20220304 Removing graph output op might cause error.
            if op.outputs[0].name in graph.outputs:
                graph.outputs.pop(op.outputs[0].name)
                graph.outputs[input_var.name] = input_var

            input_var, output_var = op.inputs[0], op.outputs[0]
            # Patch 20230103:
            # If var.source_op is DequantizeLinear, then we do not need to quantize it twice.
            if input_var.source_op is not None and input_var.source_op.type in {
                'DequantizeLinear',
                'DequantizeFloating',
            }:
                assert input_var.source_op.num_of_input == 3, 'Quantize Node Format Error, need as least 3 inputs.'
                assert isinstance(input_var.source_op, Operation)
                scale, offset = input_var.source_op.inputs[1].value, input_var.source_op.inputs[2].value

                scale_diff = torch.max(torch.abs(scale - config.scale)).item()
                zeropoint_diff = torch.max(torch.abs(offset - config.offset)).item()
                if scale_diff < 1e-4 and zeropoint_diff < 1e-1:
                    continue

            if len(output_var.dest_ops) == 1 and output_var.dest_ops[0].type in {'QuantizeLinear', 'QuantizeFloating'}:
                assert output_var.dest_ops[0].num_of_input == 3, 'Quantize Node Format Error, need as least 3 inputs.'
                assert isinstance(output_var.dest_ops[0], Operation)
                scale, offset = output_var.dest_ops[0].inputs[1].value, output_var.dest_ops[0].inputs[2].value

                scale_diff = torch.max(torch.abs(scale - config.scale)).item()
                zeropoint_diff = torch.max(torch.abs(offset - config.offset)).item()
                if scale_diff < 1e-4 and zeropoint_diff < 1e-1:
                    continue

            # insert quant & dequant op on var
            created = self.insert_quantize_node(graph=graph, var=input_var, config=config, op=op)
            self.insert_dequantize_node(graph=graph, var=created.outputs[0], config=config, op=created)

        for op in removed_activations:
            graph.remove_operation(op, keep_coherence=True)

        return graph

    def remove_duplicated_quant_op(self, graph: BaseGraph) -> BaseGraph:
        """
        Pattern:        Quant - Dequant - Quant - Dequant

        Can reduced to: Quant - Dequant

        Some time there will be more than 1 quant operation inserted with a
        single variable. This function will remove duplicated quant operation
        from variable if it is possible.

        If inserted quant operations do not share a same zeropoint and scale,
        Then there is no way to remove any one of them.

        This function has been removed since PPQ 0.6.6, and can be replaced by function convert_operation.
        """
        return graph
        interested_pairs = []
        for qt_op in graph.operations.values():
            if qt_op.type == 'QuantizeLinear':
                if len(graph.get_upstream_operations(qt_op)) != 1:
                    continue
                if graph.get_upstream_operations(qt_op)[0].type != 'DequantizeLinear':
                    continue
                interested_pairs.append((qt_op, graph.get_upstream_operations(qt_op)[0]))

        mark_to_remove = set()
        for qt_op, dq_op in interested_pairs:
            assert isinstance(dq_op, Operation) and isinstance(qt_op, Operation)
            scale_diff = torch.max(torch.abs(dq_op.inputs[1].value - qt_op.inputs[1].value)).item()
            zeropoint_diff = torch.max(torch.abs(dq_op.inputs[2].value - qt_op.inputs[2].value)).item()
            if scale_diff < 1e-5 and zeropoint_diff < 0.5:  # zero point 是整数，所以只要误差小于1就行了。
                # mark quant operation and its following operation(suppose to be another dequantization op)
                mark_to_remove.add(qt_op)
                assert len(graph.get_downstream_operations(qt_op)) == 1, 'Oops, that should never happen.'
                mark_to_remove.add(graph.get_downstream_operations(qt_op)[0])

        for op in mark_to_remove:
            assert isinstance(op, Operation)
            input_var, output_var = op.inputs[0], op.outputs[0]
            graph.remove_operation(op)
            graph.create_link_with_var(input_var, output_var)

        """
        There is another type of fusion:
        Pattern:        Quant +-- Dequant
                              |
                              +-- Dequant

        Can reduce to:  Quant - Dequant +--
                                        |
                                        +--

        Not implemented.
        """
        return graph

    @property
    def required_opsets(self) -> Dict[str, int]:
        extra_domain_versions = [('ai.onnx', 13)]
        return dict(extra_domain_versions)

    def convert_operation_from_opset11_to_opset13(self, graph: BaseGraph) -> None:
        """Convert your network from opset 11 standard towards opset 13 With
        Onnx definition, per-channel quant operation requires opset 13.

        Args:
            graph (BaseGraph): Processing graph.
        """
        # this func transform representation of certain op from opset 11 to 13
        for op in graph.operations.values():
            if op.type == 'ReduceSum' or op.type == 'Squeeze' or op.type == 'Unsqueeze':
                if 'axes' not in op.attributes:
                    continue  # is already v13
                axes = convert_any_to_torch_tensor(op.attributes.pop('axes'), dtype=torch.int64)
                graph.create_variable(name=None, value=axes, is_parameter=True, dest_ops=[op])

            elif op.type == 'Split':
                if 'split' not in op.attributes:
                    continue  # split is already v13
                split = convert_any_to_torch_tensor(op.attributes.pop('split'), dtype=torch.int64)
                graph.create_variable(name=None, value=split, is_parameter=True, dest_ops=[op])

    def convert_operation(self, graph: BaseGraph, op: QuantableOperation, quantized_param: bool):
        """Convert an operation to onnx quant & dequant format by inserting
        necessary quant & dequant op around it. There are 2 ways to represent
        quantized ONNX models:

        Operator Oriented. All the quantized operators have their own ONNX definitions,
            like QLinearConv, MatMulInteger and etc.

        Tensor Oriented, aka Quantize and DeQuantize (QDQ).
            This format uses DQ(Q(tensor)) to simulate the quantize and dequantize process,
            and QuantizeLinear and DeQuantizeLinear operators also carry the quantization parameters.

        Quantization-Aware training (QAT) models converted from Tensorflow or exported from PyTorch.

        Quantized models converted from tflite and other framework.

        Args:
            graph (BaseGraph): PPQ IR
            op (Operation): Converting op
            process_activation (bool): Converting op's activation
            process_parameter (bool): Converting op's parameter
            quantized_param (bool): Export parameter in quantized format.
        """
        # collect quantable vars, where we need to insert quant and dequant op
        for config, var in [_ for _ in op.config_with_variable]:
            inserting, inserting_var = op, var
            if not QDQHelper.TQC_Exportable_Check(TQC=config, bounded_var=var):
                continue

            if var.is_parameter:
                assert len(var.dest_ops) == 1, (
                    f'Can not export variable {var.name}, cause it has more than 1 destination operations. '
                    'PPQ require all parameters to have only 1 destination operation.'
                )

                # override quantization state, so that we can export parameter correctly.
                if config.state == QuantizationStates.BAKED:
                    config.state = QuantizationStates.ACTIVATED
                if config.state == QuantizationStates.PASSIVE_BAKED:
                    config.state = QuantizationStates.PASSIVE

                # if not quant parameter to int, all parameter should export as fp32.
                # needs insert both quant and dequant op for them
                if not quantized_param:
                    created = self.insert_quantize_node(graph=graph, var=inserting_var, config=config, op=inserting)
                    inserting_var = created.outputs[0]
                    inserting = created

                self.insert_dequantize_node(graph=graph, var=inserting_var, config=config, op=inserting)

                if quantized_param and config.policy.has_property(QuantizationProperty.LINEAR):
                    var.value = PPQLinearQuant_toInt(tensor=var.value, config=config)

            elif not var.is_parameter:
                # Patch 20230103:
                # If var.source_op is DequantizeLinear, then we do not need to quantize it twice.
                if var.source_op is not None and var.source_op.type in {'DequantizeLinear', 'DequantizeFloating'}:
                    assert var.source_op.num_of_input == 3, 'Quantize Node Format Error, need as least 3 inputs.'
                    assert isinstance(var.source_op, Operation)
                    scale, offset = var.source_op.inputs[1].value, var.source_op.inputs[2].value

                    scale_diff = torch.max(torch.abs(scale - config.scale)).item()
                    zeropoint_diff = torch.max(torch.abs(offset - config.offset)).item()
                    if scale_diff < 1e-4 and zeropoint_diff < 1e-1:
                        continue

                if len(var.dest_ops) == 1 and var.dest_ops[0].type in {'QuantizeLinear', 'QuantizeFloating'}:
                    assert var.dest_ops[0].num_of_input == 3, 'Quantize Node Format Error, need as least 3 inputs.'
                    assert isinstance(var.dest_ops[0], Operation)
                    scale, offset = var.dest_ops[0].inputs[1].value, var.dest_ops[0].inputs[2].value

                    scale_diff = torch.max(torch.abs(scale - config.scale)).item()
                    zeropoint_diff = torch.max(torch.abs(offset - config.offset)).item()
                    if scale_diff < 1e-4 and zeropoint_diff < 1e-1:
                        continue

                created = self.insert_quantize_node(graph=graph, var=inserting_var, config=config, op=inserting)
                inserting_var = created.outputs[0]
                inserting = created

                self.insert_dequantize_node(graph=graph, var=inserting_var, config=config, op=inserting)

    def prepare_graph(
        self, graph: BaseGraph, remove_activation_fn: bool = True, quant_parameter_to_int: bool = True
    ) -> BaseGraph:
        """Prepare your graph for exporting.

        There are many works to do with your graph:

            1. Insert Quant and Dequant operation within your graph.

            2. Remove all unnecessary activations.

            3. Quantize all parameters of your graph, convert them to int8.

        Args:
            graph (BaseGraph): Processing Graph

        Returns:
            BaseGraph: Processed Graph
        """
        self.convert_operation_from_opset11_to_opset13(graph)

        # mark quantable variables
        for op in graph.topological_sort():
            if not isinstance(op, QuantableOperation):
                continue
            if op.type in {'QuantizeLinear', 'DequantizeLinear', 'QuantizeFloating', 'DequantizeFloating'}:
                continue

            self.convert_operation(graph=graph, op=op, quantized_param=quant_parameter_to_int)

        # remove activations
        if remove_activation_fn:
            # remove useless activation.
            self.remove_activation_ops(graph)

        return self.remove_duplicated_quant_op(graph)

    def export(
        self,
        file_path: str,
        graph: BaseGraph,
        config_path: str = None,
        quantized_param: bool = True,
        remove_activation: bool = True,
        save_as_external_data: bool = False,
    ) -> None:
        """
        Export PPQ Graph to Onnx QDQ format.
            This function requires a set of parameters to configure onnx format.

        Args:
            file_path (str): Onnx file name.

            graph (BaseGraph): Exporting ppq graph.

            config_path (str, optional): config file is a json file that contains quant-related
                information, this file is require by TensorRT for initialize its quantization
                pipeline. If config_path = None, no json file will be created.

            export_QDQ_op (bool, optional): whether to export QDQ node in onnx model.

            quantized_param (bool, optional): export quantized parameter, if quantized_param = False,
                PPQ will export parameter in FP32 format.

            remove_activation (bool, optional): this option will remove activation op(Relu, Clip),
                requires ASYMMTRICAL quantizaiton.

            save_as_external_data (bool, optional): for model larger than 2GB,
                this option will split model into external param files.
        """
        # In prepare stage, quant & dequant node are inserted into graph.
        graph = self.prepare_graph(
            graph, remove_activation_fn=remove_activation, quant_parameter_to_int=quantized_param
        )

        # if a valid config path is given, export quantization config to there.
        if config_path is not None:
            super().export_quantization_config(config_path, graph)

        # before we can export them, we firstly convert all ops to proper format.
        for op in [_ for _ in graph.topological_sort()]:
            if op.type in OP_CONVERTERS:
                exporter = OP_CONVERTERS[op.type]()
                assert isinstance(exporter, OperationExporter), (
                    f'Expected an OpExporter here, however {type(exporter)} was given.'
                )
                op = exporter.export(op=op, graph=graph)

        name = graph._name
        if not name:
            name = 'PPL Quantization Tool - Onnx Export'

        # Ready to export onnx graph definition.
        _inputs, _outputs, _initilizers, _nodes, _value_info = [], [], [], [], []
        for operation in graph.topological_sort():
            _nodes.append(super().build_operator_proto(operation))

        for variable in graph.variables.values():
            tensor_proto = super().build_variable_proto(variable)
            if variable.name in graph.inputs:
                _inputs.append(tensor_proto)
            if variable.name in graph.outputs:
                _outputs.append(tensor_proto)
            if variable.is_parameter:
                _initilizers.append(tensor_proto)
            else:
                _value_info.append(tensor_proto)

        graph_def = helper.make_graph(
            name=name, nodes=_nodes, inputs=_inputs, outputs=_outputs, initializer=_initilizers, value_info=_value_info
        )
        extra_opsets = self.required_opsets

        opsets = []
        if GRAPH_OPSET_ATTRIB in graph._detail:
            for opset in graph._detail[GRAPH_OPSET_ATTRIB]:
                if opset['domain'] in extra_opsets or opset['domain'] == '':
                    continue
                op = onnx.OperatorSetIdProto()
                op.domain = opset['domain']
                op.version = opset['version']
                opsets.append(op)

        for key, value in extra_opsets.items():
            op = onnx.OperatorSetIdProto()
            # PATCH 2024.01.30
            # 我也不知道为什么 onnx checker 会对 ai.onnx 这个 domain 报错
            # 按照规范 ai.onnx 与 "" 是等价的写法
            if key == 'ai.onnx':
                key = ""
            op.domain = key
            op.version = value
            opsets.append(op)

        onnx_model = helper.make_model(graph_def, producer_name=PPQ_CONFIG.NAME, opset_imports=opsets)
        onnx_model.ir_version = 7
        # onnx.checker.check_model(onnx_model)
        size_threshold = 0 if save_as_external_data else 1024
        onnx.save(
            onnx_model,
            file_path,
            size_threshold=size_threshold,
            save_as_external_data=save_as_external_data,
            all_tensors_to_one_file=(not save_as_external_data),
        )

        # Check Graph
        unsupportable_quant_op = set()
        for op in graph.operations.values():
            if isinstance(op, QuantableOperation):
                for cfg, var in op.config_with_variable:
                    if not QDQHelper.TQC_Exportable_Check(TQC=cfg, bounded_var=var):
                        continue
                    if cfg.num_of_bits != 8 or cfg.policy.has_property(QuantizationProperty.FLOATING):
                        unsupportable_quant_op.add(op)

        if len(unsupportable_quant_op) != 0:
            ppq_warning('Exported Onnx Model is not executable, following Op has onnxruntime-unsupported quant policy:')
            for op in unsupportable_quant_op:
                ppq_warning(f'{op.name} (bitwidth != 8)')
            ppq_warning(
                'For Generating onnxruntime-executable Model, use TargetPlatform = Onnxruntime or OnnxruntimeQuantizer instead.'
            )
